from ..person import Person
from django import forms
from django.forms import DateInput
from ...patient import PatientExamination
from ...data_file import ReportFile
from django.db import models
from faker import Faker
import random
from datetime import datetime
from icecream import ic
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from endoreg_db.models import Center, Gender


class Patient(Person):
    """
    A class representing a patient.

    Attributes inhereted from Person:
        first_name (str): The first name of the patient.
        last_name (str): The last name of the patient.
        dob (datetime.date): The date of birth of the patient.
        gender (Foreign Key): The gender of the patient.
        email (str): The email address of the patient.
        phone (str): The phone number of the patient.

    """

    # -----gc-08-dev--changings---
    first_name = models.CharField(max_length=100)
    last_name = models.CharField(max_length=100)
    dob = models.DateField(null=True, blank=True)
    gender = models.ForeignKey(
        "Gender", on_delete=models.SET_NULL, null=True, blank=True
    )
    center = models.ForeignKey(
        "Center", on_delete=models.SET_NULL, null=True, blank=True
    )
    patient_hash = models.CharField(max_length=255, blank=True, null=True)

    def __str__(self):
        return f"{self.first_name} {self.last_name} ({self.dob})"

    @classmethod
    def get_or_create_pseudo_patient_by_hash(
        cls,
        patient_hash: str,
        center: "Center" = None,
        gender: "Gender" = None,
        birth_month: int = None,
        birth_year: int = None,
    ):
        from endoreg_db.utils import random_day_by_year, create_mock_patient_name

        created = False

        existing_pathient = cls.objects.filter(patient_hash=patient_hash).first()
        if existing_pathient:
            ic(f"Patient with hash {patient_hash} already exists")
            ic(f"Returning existing patient: {existing_pathient}")
            return existing_pathient, created

        # If no patient with the given hash exists, create a new pseudo patient
        assert center, "Center must be provided to create a new pseudo patient"
        assert gender, "Gender must be provided to create a new pseudo patient"
        assert birth_month, (
            "Birth month must be provided to create a new pseudo patient"
        )
        assert birth_year, "Birth year must be provided to create a new pseudo patient"

        pseudo_dob = random_day_by_year(birth_year)
        gender_name = gender.name
        first_name, last_name = create_mock_patient_name(gender_name)

        ic(f"Creating pseudo patient with hash {patient_hash}")
        ic(f"Generated name: {first_name} {last_name}")

        patient = cls.objects.create(
            first_name=first_name,
            last_name=last_name,
            dob=pseudo_dob,
            patient_hash=patient_hash,
            is_real_person=False,
        )

        patient.save()
        created = True

        return patient, created

    def export_patient_examinations(self):
        """
        Get all associated PatientExaminations, ReportFiles, and Videos for the patient.
        """
        from endoreg_db.models import PatientExamination, ReportFile, Video

        patient_examinations = PatientExamination.objects.filter(patient=self)
        report_files, videos = [], []
        for patient_examination in patient_examinations:
            rr = patient_examination.reportfile_set.all()
            vv = patient_examination.videos.all()

            report_files.extend(rr)
            videos.extend(vv)

        return patient_examinations, report_files, videos

    def get_dob(self) -> datetime.date:
        dob: datetime.date = self.dob
        return dob

    def get_unmatched_report_files(
        self,
    ):  # field: self.report_files; filter: report_file.patient_examination = None
        """Returns all report files for this patient that are not matched to a patient examination."""

        return self.reportfile_set.filter(patient_examination=None)

    def get_unmatched_video_files(
        self,
    ):  # field: self.videos; filter: video.patient_examination = None
        """Returns all video files for this patient that are not matched to a patient examination."""
        return self.videos.filter(patient_examination=None)

    def get_patient_examinations(self):  # field: self.patient_examinations
        """Returns all patient examinations for this patient ordered by date (most recent is first)."""
        return self.patient_examinations.order_by("-date")

    def create_examination(
        self,
        examination_name_str: str = None,
        date_start: datetime = None,
        date_end: datetime = None,
    ):
        """Creates a patient examination for this patient."""

        if examination_name_str:
            from endoreg_db.models import Examination

            examination = Examination.objects.get(name=examination_name_str)
            patient_examination = PatientExamination(
                patient=self,
                examination=examination,
                date_start=date_start,
                date_end=date_end,
            )

        else:
            patient_examination = PatientExamination(
                patient=self, date_start=date_start, date_end=date_end
            )

        patient_examination.save()

        return patient_examination

    def create_examination_by_indication(
        self, indication, date_start: datetime = None, date_end: datetime = None
    ):
        from endoreg_db.models import (
            ExaminationIndication,
            Examination,
            PatientExaminationIndication,
        )

        assert isinstance(indication, ExaminationIndication)

        examination = indication.get_examination()

        assert isinstance(examination, Examination)

        patient_examination = PatientExamination.objects.create(
            patient=self,
            examination=examination,
            date_start=date_start,
            date_end=date_end,
        )

        patient_examination.save()

        patient_examination_indication = PatientExaminationIndication.objects.create(
            patient_examination=patient_examination, examination_indication=indication
        )
        patient_examination_indication.save()

        return patient_examination, patient_examination_indication

    def create_event(
        self,
        event_name_str: str,
        date_start: datetime = None,
        date_end: datetime = None,
        description: str = None,
    ):
        from endoreg_db.models import Event, PatientEvent

        event = Event.objects.get(name=event_name_str)

        if not date_start:
            date_start = datetime.now()

        patient_event = PatientEvent.objects.create(
            patient=self,
            event=event,
            date_start=date_start,
        )

        return patient_event

    def create_examination_by_report_file(self, report_file: ReportFile):
        """Creates a patient examination for this patient based on the given report file."""
        patient_examination = PatientExamination(patient=self, report_file=report_file)
        patient_examination.save()
        return patient_examination

    @classmethod
    def get_random_gender(cls, p_male=0.5, p_female=0.5):
        """
        Get a Gender object by name (male, female) from the database with given probability.

        :param p_male: Probability of selecting 'male' gender.
        :param p_female: Probability of selecting 'female' gender.
        :return: Gender object selected based on given probabilities.
        """
        from endoreg_db.models import Gender

        # Extract names and probabilities
        gender_names = ["male", "female"]
        probabilities = [p_male, p_female]

        # Debug: print the names and probabilities
        # print(f"Gender names: {gender_names}")
        # print(f"Probabilities: {probabilities}")

        # Select a gender based on the given probabilities
        selected_gender = random.choices(gender_names, probabilities)[0]
        # Debug: print the selected gender
        # print(f"Selected gender: {selected_gender}")

        # Fetch the corresponding Gender object from the database
        gender_obj = Gender.objects.get(name=selected_gender)

        return gender_obj

    @classmethod
    def get_random_age(
        cls, min_age=55, max_age=90, mean_age=65, std_age=10, distribution="normal"
    ):
        """
        Get a random age based on the given distribution.

        :param min_age: Minimum age.
        :param max_age: Maximum age.
        :param mean_age: Mean age.
        :param std_age: Standard deviation of the age.
        :param distribution: Distribution of the age.
        :return: Random age based on the given distribution.
        """
        if distribution == "normal":
            age = int(random.normalvariate(mean_age, std_age))
        else:
            age = int(random.uniform(min_age, max_age))

        return age

    @classmethod
    def get_dob_from_age(cls, age, current_date=None):
        """
        Get a date of birth based on the given age and current date.

        :param age: Age of the patient.
        :param current_date: Current date.
        :return: Date of birth based on the given age and current date.
        """
        if current_date is None:
            current_date = datetime.now()
        dob = current_date.replace(year=current_date.year - age).date()

        # TODO
        # randomize the day and month by adding a random number of days (0-364) to the date

        return dob

    @classmethod
    def get_random_name_for_gender(cls, gender_obj, locale="de_DE"):
        gender = gender_obj.name
        fake = Faker(locale)

        if gender == "male":
            first_name = fake.first_name_male()
            last_name = fake.last_name_male()

        else:
            first_name = fake.first_name_female()
            last_name = fake.last_name_female()

        return last_name, first_name

    @classmethod
    def create_generic(cls, center="gplay_case_generator"):
        """
        Create a generic patient with random attributes.

        :param center: The center of the patient.
        :return: The created patient.
        """
        from endoreg_db.models import Center

        gender = Patient.get_random_gender()
        last_name, first_name = Patient.get_random_name_for_gender(gender)

        age = Patient.get_random_age()
        dob = Patient.get_dob_from_age(age)

        center = Center.objects.get(name=center)

        patient = Patient.objects.create(
            first_name=first_name,
            last_name=last_name,
            dob=dob,
            gender=gender,
        )
        patient.save()
        return patient

    def age(self):
        """
        Get the age of the patient.

        :return: The age of the patient.
        """
        # calculate correct age based on current date including day and month
        current_date = datetime.now()
        dob = self.dob
        age = (
            current_date.year
            - dob.year
            - ((current_date.month, current_date.day) < (dob.month, dob.day))
        )
        return age

    def create_lab_sample(self, sample_type="generic", date=None, save=True):
        """
        Create a lab sample for this patient.

        :param sample_type: The sample type. Should be either string of the sample types
            name or the sample type object. If not set, the default sample type ("generic") is used.
        :param date: The date of the lab sample.
        :return: The created lab sample.
        """
        from endoreg_db.models import PatientLabSample, PatientLabSampleType

        if date is None:
            date = datetime.now()

        if isinstance(sample_type, str):
            sample_type = PatientLabSampleType.objects.get(name=sample_type)
            assert sample_type is not None, (
                f"Sample type with name '{sample_type}' not found."
            )  #
        elif not isinstance(sample_type, PatientLabSampleType):
            raise ValueError(
                "Sample type must be either a string or a PatientLabSampleType object."
            )

        patient_lab_sample = PatientLabSample.objects.create(
            patient=self, sample_type=sample_type, date=date
        )

        if save:
            patient_lab_sample.save()

        return patient_lab_sample


class PatientForm(forms.ModelForm):
    class Meta:
        model = Patient
        fields = "__all__"
        widgets = {
            "dob": DateInput(attrs={"type": "date"}),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.fields.values():
            field.widget.attrs["class"] = "form-control"



from django.core.management.base import BaseCommand
from endoreg_db.models import (
    CaseTemplateRuleType,
    CaseTemplateRuleValueType,
    CaseTemplateRuleValue,
    CaseTemplateRule,
    CaseTemplateType,
    CaseTemplate,
    SingleCategoricalValueDistribution,
    NumericValueDistribution,
    MultipleCategoricalValueDistribution,
    DateValueDistribution,

    QuizQuestionType,
    QuizAnswerType,
    QuizQuestion,
    QuizAnswer,

    # Other models for ForeignKeys
    Unit, Center
)
from collections import OrderedDict

from ...utils import load_model_data_from_yaml
from ...data import (
    CASE_TEMPLATE_RULE_TYPE_DATA_DIR,
    CASE_TEMPLATE_RULE_DATA_DIR,
    CASE_TEMPLATE_RULE_VALUE_TYPE_DATA_DIR,
    CASE_TEMPLATE_RULE_VALUE_DATA_DIR,
    CASE_TEMPLATE_TYPE_DATA_DIR,
    CASE_TEMPLATE_DATA_DIR,
)

IMPORT_METADATA = OrderedDict({
    CaseTemplateRuleType.__name__: {
        "dir": CASE_TEMPLATE_RULE_TYPE_DATA_DIR,
        "model": CaseTemplateRuleType, 
        "foreign_keys": [], 
        "foreign_key_models": [] 
    },
    CaseTemplateRuleValueType.__name__: {
        "dir": CASE_TEMPLATE_RULE_VALUE_TYPE_DATA_DIR,
        "model": CaseTemplateRuleValueType, 
        "foreign_keys": [], 
        "foreign_key_models": [] 
    },
    CaseTemplateRuleValue.__name__: {
        "dir": CASE_TEMPLATE_RULE_VALUE_DATA_DIR,
        "model": CaseTemplateRuleValue, 
        "foreign_keys": ["value_type"], 
        "foreign_key_models": [CaseTemplateRuleValueType] 
    },
    CaseTemplateRule.__name__: {
        "dir": CASE_TEMPLATE_RULE_DATA_DIR,
        "model": CaseTemplateRule, 
        "foreign_keys": [
            "rule_type", 
            # "rule_values",
            "chained_rules", 
            "value_type",
            "single_categorical_value_distribution",
            "numerical_value_distribution",
            "multiple_categorical_value_distribution",
            "date_value_distribution"
        ], 
        "foreign_key_models": [
            CaseTemplateRuleType,
            # CaseTemplateRuleValue,
            CaseTemplateRule,
            CaseTemplateRuleValueType,
            SingleCategoricalValueDistribution,
            NumericValueDistribution,
            MultipleCategoricalValueDistribution,
            DateValueDistribution

        ] 
    },
    CaseTemplateType.__name__: {
        "dir": CASE_TEMPLATE_TYPE_DATA_DIR,
        "model": CaseTemplateType, 
        "foreign_keys": [], 
        "foreign_key_models": [] 
    },
    CaseTemplate.__name__: {
        "dir": CASE_TEMPLATE_DATA_DIR,
        "model": CaseTemplate, 
        "foreign_keys": ["template_type", "rules"], 
        "foreign_key_models": [CaseTemplateType, CaseTemplateRule] 
    },
    
})

class Command(BaseCommand):
    help = """Load all .yaml files in the data/intervention directory
    into the Intervention and InterventionType model"""

    def add_arguments(self, parser):
        parser.add_argument(
            '--verbose',
            action='store_true',
            help='Display verbose output',
        )

    def handle(self, *args, **options):
        verbose = options['verbose']
        for model_name in IMPORT_METADATA.keys():
            _metadata = IMPORT_METADATA[model_name]
            load_model_data_from_yaml(
                self,
                model_name,
                _metadata,
                verbose
            )


"""
This module defines the ModelMeta and ModelMetaManager classes for managing AI model metadata.
"""

import os
from pathlib import Path
from typing import Optional, TYPE_CHECKING
import shutil

from django.db import models
from django.core.validators import FileExtensionValidator
from icecream import ic

if TYPE_CHECKING:
    from endoreg_db.models import LabelSet, AiModel  # pylint: disable=import-outside-toplevel

PSEUDO_DIR = Path(os.environ.get("DJANGO_PSEUDO_DIR", Path("./erc_data")))

STORAGE_LOCATION = PSEUDO_DIR
WEIGHTS_DIR_NAME = "db_model_weights"
WEIGHTS_DIR = STORAGE_LOCATION / WEIGHTS_DIR_NAME

if not WEIGHTS_DIR.exists():
    WEIGHTS_DIR.mkdir(parents=True)


class ModelMetaManager(models.Manager):
    """
    Custom manager for ModelMeta with additional query methods.
    """

    def get_by_natural_key(self, name: str, version: str) -> "ModelMeta":
        """Get the model by its natural key"""
        return self.get(name=name, version=version)


class ModelMeta(models.Model):
    """
    Attributes:
        name (str): The name of the model.
        version (str): The version of the model.
        type (ModelType): The type of the model.
        labelset (LabelSet): Associated labels.
        weights (FileField): Path to the model weights.
    """

    name = models.CharField(max_length=255)
    version = models.CharField(max_length=255)

    model = models.ForeignKey(
        "AiModel", on_delete=models.CASCADE, related_name="models"
    )

    labelset = models.ForeignKey(
        "LabelSet", on_delete=models.CASCADE, related_name="models"
    )
    activation = models.CharField(max_length=255, default="sigmoid")

    weights = models.FileField(
        upload_to=WEIGHTS_DIR_NAME,
        validators=[FileExtensionValidator(allowed_extensions=["ckpt"])],
        # storage=FileSystemStorage(location=STORAGE_LOCATION.resolve().as_posix()),
        null=True,
        blank=True,
    )

    # Means and stds for normalization
    mean = models.CharField(max_length=255, default="0.45211223,0.27139644,0.19264949")
    std = models.CharField(max_length=255, default="0.31418097,0.21088019,0.16059452")

    size_x = models.IntegerField(default=716)
    size_y = models.IntegerField(default=716)
    axes = models.CharField(max_length=255, default="2,0,1")
    batchsize = models.IntegerField(default=16)
    num_workers = models.IntegerField(default=0)

    description = models.TextField(blank=True, null=True)
    date_created = models.DateTimeField(auto_now_add=True)

    objects = ModelMetaManager()

    if TYPE_CHECKING:
        labelset: "LabelSet"
        ai_model: "AiModel"

    @classmethod
    def create_from_file(
        cls,
        meta_name,
        model_name,
        model_meta_version,
        labelset_name,
        weights_file,
        bump_version,
        **kwargs,
    ):
        """Create a new model meta from a file"""

        from endoreg_db.models import LabelSet, AiModel  # pylint: disable=import-outside-toplevel

        ic("Start")
        model_meta_version = kwargs.get("model_meta_version", None)

        ai_model = AiModel.objects.get(name=model_name)

        # check If ModelMeta with same name and model already exists
        if not model_meta_version:
            if cls.objects.filter(name=meta_name, model=ai_model).exists():
                # get highest version number
                highest_version = (
                    cls.objects.filter(name=meta_name, model=ai_model)
                    .latest("version")
                    .version
                )
                version = int(highest_version) + 1
            else:
                version = 1
        else:
            version = model_meta_version

        meta_exists = cls.objects.filter(
            name=meta_name, model=ai_model, version=version
        ).exists()
        if meta_exists and not bump_version:
            raise ValueError(
                f"ModelMeta with name {meta_name} and version {version} already exists"
            )

        assert labelset_name is not None, "Labelset name must be provided"
        labelset = LabelSet.objects.get(name=labelset_name)
        assert labelset is not None, "Labelset not found"

        weights_path = Path(weights_file)
        # If not under our WEIGHTS_DIR, copy it there
        if not str(weights_path).startswith(str(WEIGHTS_DIR)):
            target_path = WEIGHTS_DIR / weights_path.name
            if not target_path.exists():
                shutil.copy(weights_path, target_path)
            weights_file = target_path

        return cls.objects.create(
            name=meta_name,
            model=ai_model,
            version=version,
            labelset=labelset,
            weights=str(weights_file),
            **kwargs,
        )

    @classmethod
    def get_latest_version(cls, name) -> int:
        """
        get the model meta with the highest version. Assumes Model Name and Meta Name are the same
        """
        from endoreg_db.models import AiModel  # pylint: disable=import-outside-toplevel

        ai_model = AiModel.objects.get(name=name)
        ic(f"Model: {ai_model}")
        assert ai_model is not None, "Model not found"

        model_meta = cls.objects.filter(name=name, model=ai_model).latest("version")
        ic(f"Latest ModelMeta: {model_meta}")

        latest_version = model_meta.version
        ic(f"Latest Version: {latest_version}")

        return int(latest_version)

    @classmethod
    def get_activation_function(cls, activation: str):
        """get the activation function"""
        from torch import nn  # pylint: disable=import-outside-toplevel

        lookup = {
            "sigmoid": nn.Sigmoid(),
            "softmax": nn.Softmax(),
            "none": nn.Identity(),
        }

        return lookup[activation]

    def get_inference_dataset_config(self):
        """
        Creates a dictionary with the configuration for the inference dataset.

        Example:
            sample_config = {
                # mean and std for normalization
                "mean": (0.45211223, 0.27139644, 0.19264949),
                "std": (0.31418097, 0.21088019, 0.16059452),
                # Image Size
                "size_x": 716,
                "size_y": 716,
                # how to wrangle axes of the image before putting them in the network
                "axes": [2,0,1],  # 2,1,0 for opencv
                "batchsize": 16,
                "num_workers": 0, # always 1 for Windows systems # FIXME: fix celery crash if multiprocessing
                # maybe add sigmoid after prediction?
                "activation": nn.Sigmoid(),
                "labels": ['appendix',  'blood',  'diverticule',  'grasper',  'ileocaecalvalve',  'ileum',  'low_quality',  'nbi',  'needle',  'outside',  'polyp',  'snare',  'water_jet',  'wound']
            }
        """
        mean = list(map(float, str(self.mean).split(",")))
        std = list(map(float, str(self.std).split(",")))
        axes = list(map(int, str(self.axes).split(",")))

        config = {
            "mean": mean,
            "std": std,
            "size_x": self.size_x,
            "size_y": self.size_y,
            "axes": axes,
            "batchsize": self.batchsize,
            "num_workers": self.num_workers,
            "activation": self.get_activation_function(self.activation),
            "labels": self.labelset.get_labels_in_order(),  # pylint: disable=no-member
        }

        ic(f"ModelMeta - get_inference_dataset_config: {config}")

        return config

    def natural_key(self):
        """Return the natural key for this model"""
        return (self.name, self.version)

    def __str__(self):
        return f"{self.name} (v: {self.version})"

    def get_config_dict(self) -> dict:
        """Return the configuration as a dictionary"""
        conf_dict = {
            "mean": list(map(float, str(self.mean).split(","))),
            "std": list(map(float, str(self.std).split(","))),
            "size_x": self.size_x,
            "size_y": self.size_y,
            "axes": list(map(int, str(self.axes).split(","))),
            "batchsize": self.batchsize,
            "num_workers": self.num_workers,
            "activation": self.activation,
            "labels": self.labelset.get_labels_in_order(),  # pylint: disable=no-member
        }
        return conf_dict

    @classmethod
    def get_by_name(cls, name: str, version: Optional[int]) -> "ModelMeta":
        """Get the model by its name and version"""
        if version is None:
            version = cls.get_latest_version(name)
        return cls.objects.get(name=name, version=version)



from rest_framework import serializers
from endoreg_db.models import (
    Center,
)

class CenterSerializer(serializers.ModelSerializer):
    class Meta:
        model = Center
        fields = '__all__'





from django.core.management.base import BaseCommand
from endoreg_db.models import (
    Finding, FindingType, Examination,
    FindingLocationClassification, FindingLocationClassificationChoice,
    Organ,
    FindingMorphologyClassificationType, FindingMorphologyClassification,
    FindingMorphologyClassificationChoice,
    FindingIntervention, FindingInterventionType, LabValue, Contraindication

)
from ...utils import load_model_data_from_yaml
from ...data import (
    FINDING_DATA_DIR, FINDING_TYPE_DATA_DIR,
    FINDING_LOCATION_CLASSIFICATION_DATA_DIR, 
    FINDING_LOCATION_CLASSIFICATION_CHOICE_DATA_DIR,
    FINDING_MORPHOLOGY_CLASSIFICATION_CHOICE_DATA_DIR, 
    FINDING_MORPHOLOGY_CLASSIFICATION_DATA_DIR,
    FINDING_MORPGOLOGY_CLASSIFICATION_TYPE_DATA_DIR,
    FINDING_INTERVETION_DATA_DIR,
    FINIDNG_INTERVENTION_TYPE_DATA_DIR
)

MODEL_0 = FindingType
MODEL_1 = Finding
MODEL_2 = FindingLocationClassificationChoice
MODEL_3 = FindingLocationClassification
MODEL_4 = FindingMorphologyClassificationType
MODEL_5 = FindingMorphologyClassification
MODEL_6 = FindingMorphologyClassificationChoice
MODEL_7 = FindingInterventionType
MODEL_8 = FindingIntervention

IMPORT_MODELS = [ # string as model key, serves as key in IMPORT_METADATA
    MODEL_7.__name__,
    MODEL_8.__name__,
    MODEL_0.__name__,
    MODEL_1.__name__,
    MODEL_2.__name__,
    MODEL_3.__name__,
    MODEL_4.__name__,
    MODEL_5.__name__,
    MODEL_6.__name__,
]

IMPORT_METADATA = {
    MODEL_0.__name__: {
        "dir": FINDING_TYPE_DATA_DIR, # e.g. "interventions"
        "model": MODEL_0, # e.g. Intervention
        "foreign_keys": [], # e.g. ["intervention_types"]
        "foreign_key_models": [] # e.g. [InterventionType]
    },
    Finding.__name__: {
        "dir": FINDING_DATA_DIR, # e.g. "interventions"
        "model": MODEL_1, # e.g. Intervention
        "foreign_keys": [
            "finding_types",
            "examinations",
            "finding_interventions",
            "causing_finding_interventions",
            "opt_causing_finding_interventions",
        ], # e.g. ["intervention_types"]
        "foreign_key_models": [
            FindingType,
            Examination,
            FindingIntervention,
            FindingIntervention,
            FindingIntervention,
        ] # e.g. [InterventionType]
    },
    MODEL_2.__name__: {
        "dir": FINDING_LOCATION_CLASSIFICATION_CHOICE_DATA_DIR, # e.g. "interventions"
        "model": MODEL_2, # e.g. Intervention
        "foreign_keys": [
            "organs"
        ], # e.g. ["intervention_types"]
        "foreign_key_models": [
            Organ
        ] # e.g. [InterventionType]
    },
    MODEL_3.__name__: {
        "dir": FINDING_LOCATION_CLASSIFICATION_DATA_DIR, # e.g. "interventions"
        "model": MODEL_3, # e.g. Intervention
        "foreign_keys": [
            "examinations",
            "finding_types",
            "findings",
            "choices"
        ], # e.g. ["intervention_types"]
        "foreign_key_models": [
            Examination,
            FindingType,
            Finding,
            FindingLocationClassificationChoice
        ] # e.g. [InterventionType]
    },
    MODEL_4.__name__: {
        "dir": FINDING_MORPGOLOGY_CLASSIFICATION_TYPE_DATA_DIR, # e.g. "interventions"
        "model": MODEL_4, # e.g. Intervention
        "foreign_keys": [
            "required_by_findings",
            "optional_for_findings"
        ], # e.g. ["intervention_types"]
        "foreign_key_models": [
            Finding,
            Finding
        ] # e.g. [InterventionType]
    },

    MODEL_5.__name__: {
        "dir": FINDING_MORPHOLOGY_CLASSIFICATION_DATA_DIR, # e.g. "interventions"
        "model": MODEL_5, # e.g. Intervention
        "foreign_keys": [
            "classification_type",
        ],
        "foreign_key_models": [
            FindingMorphologyClassificationType,
        ] 
    },
    MODEL_6.__name__: {
        "dir": FINDING_MORPHOLOGY_CLASSIFICATION_CHOICE_DATA_DIR, # e.g. "interventions"
        "model": MODEL_6, # e.g. Intervention
        "foreign_keys": [
            "classification",
        ],
        "foreign_key_models": [
            FindingMorphologyClassification,
        ] 
    },
    FindingInterventionType.__name__: {
        "dir": FINIDNG_INTERVENTION_TYPE_DATA_DIR, # e.g. "interventions"
        "model": FindingInterventionType, # e.g. Intervention
        "foreign_keys": [], # e.g. ["intervention_types"]
        "foreign_key_models": [] # e.g. [InterventionType]
    },
    FindingIntervention.__name__: {
        "dir": FINDING_INTERVETION_DATA_DIR, # e.g. "interventions"
        "model": FindingIntervention, # e.g. Intervention
        "foreign_keys": [
            "intervention_types",
            "required_lab_values",
            "contraindications"
        ], # e.g. ["intervention_types"]
        "foreign_key_models": [
            FindingInterventionType,
            LabValue,
            Contraindication
        ] # e.g. [InterventionType]
    },
}

class Command(BaseCommand):
    help = """Load all .yaml files in the data/intervention directory
    into the Intervention and InterventionType model"""

    def add_arguments(self, parser):
        parser.add_argument(
            '--verbose',
            action='store_true',
            help='Display verbose output',
        )

    def handle(self, *args, **options):
        verbose = options['verbose']
        for model_name in IMPORT_MODELS:
            _metadata = IMPORT_METADATA[model_name]
            load_model_data_from_yaml(
                self,
                model_name,
                _metadata,
                verbose
            )


from rest_framework import serializers
from endoreg_db.models import  BinaryClassificationAnnotationTask, ImageClassificationAnnotation



class BinaryClassificationAnnotationTaskSerializer(serializers.ModelSerializer):
    class Meta:
        model = BinaryClassificationAnnotationTask
        fields = '__all__'

class ImageClassificationAnnotationSerializer(serializers.ModelSerializer):
    class Meta:
        model = ImageClassificationAnnotation
        fields = '__all__'



from django.db import models

# models.py in your main app
from django.contrib.auth.models import User

class ProfessionManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class Profession(models.Model):
    objects = ProfessionManager()
    name = models.CharField(max_length=100)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)
    description = models.TextField(blank=True, null=True)

    def __str__(self):
        return self.name_de

class PortalUserInfo(models.Model):
    user = models.OneToOneField(User, on_delete=models.CASCADE)
    profession = models.ForeignKey('Profession', on_delete=models.CASCADE, blank=True, null=True)
    works_in_endoscopy = models.BooleanField(blank=True, null=True)
    # Add other fields as needed

    def __str__(self):
        return self.user.username



from django.db import models
from rest_framework import serializers

class ExaminationTimeManager(models.Manager):
    """
    Manager for ExaminationTime with custom query methods.
    """
    def get_by_natural_key(self, name: str) -> "ExaminationTime":
        return self.get(name=name)

class ExaminationTime(models.Model):
    """
    Represents a specific time configuration for examinations.

    Attributes:
        name (str): The unique name of the examination time.
        name_de (str): The German name of the examination time.
        name_en (str): The English name of the examination time.
        start_time (TimeField): The starting time for the examination.
        end_time (TimeField): The ending time for the examination.
        time_types (ManyToManyField): The types associated with this examination time.
    """
    name = models.CharField(max_length=100, unique=True)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)
    start_time = models.TimeField(blank=True, null=True)
    time_types = models.ManyToManyField('ExaminationTimeType', blank=True)
    end_time = models.TimeField(blank=True, null=True)
    objects = ExaminationTimeManager()

    def __str__(self) -> str:
        """
        String representation of the examination time.

        Returns:
            str: The name of the examination time.
        """
        return self.name

    def natural_key(self) -> tuple:
        """
        Returns the natural key for the examination time.

        Returns:
            tuple: The natural key consisting of the name.
        """
        return (self.name,)

    class Meta:
        verbose_name = 'Examination Time'
        verbose_name_plural = 'Examination Times'
        ordering = ['name']




from django.db import models


class LabelTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class LabelType(models.Model):
    """
    A class representing a label type.

    Attributes:
        name (str): The name of the label type.
        description (str): A description of the label type.

    """

    name = models.CharField(max_length=255)
    description = models.TextField(blank=True, null=True)

    objects = LabelTypeManager()

    def natural_key(self):
        """Return the natural key of this label type"""
        return (self.name,)

    def __str__(self):
        return str(self.name)


class LabelManager(models.Manager):
    """Manager class for handling Label model operations."""

    def get_by_natural_key(self, name):
        """Retrieves a Label instance by its natural key (name)."""
        return self.get(name=name)


class Label(models.Model):
    """
    A class representing a label.

    Attributes:
        name (str): The name of the label.
        label_type (LabelType): The type of the label.
        description (str): A description of the label.

    """

    name = models.CharField(max_length=255)
    label_type = models.ForeignKey(
        "LabelType", on_delete=models.CASCADE, related_name="labels"
    )
    description = models.TextField(blank=True, null=True)

    objects = LabelManager()

    def natural_key(self):
        """Return the natural key of this label"""
        return (self.name,)

    def __str__(self):
        return str(self.name)


class LabelSetManager(models.Manager):
    """
    Manager class for handling LabelSet model operations.
    Methods
    -------
    get_by_natural_key(name)

    """

    def get_by_natural_key(self, name):
        """Retrieves a LabelSet instance by its natural key (name)."""
        return self.get(name=name)


class LabelSet(models.Model):
    """
    A class representing a label set.

    Attributes:
        name (str): The name of the label set.
        description (str): A description of the label set.

    """

    name = models.CharField(max_length=255)
    description = models.TextField(blank=True, null=True)
    version = models.IntegerField()
    labels = models.ManyToManyField("Label", related_name="labels")

    objects = LabelSetManager()

    def natural_key(self):
        """Return the natural key of this label set"""
        return (self.name,)

    def __str__(self):
        return str(self.name)

    def get_labels_in_order(self):
        """
        Get all labels in this label set as list in the correct order.
        Ordered by string representation (a is first).
        """
        labels = list(self.labels.all())
        labels.sort(key=lambda x: x.name)
        return labels



from endoreg_db.models import (
    Patient, 
    Center, 
    PatientLabValue, LabValue
)

from django.core.management import call_command
from django.test import TestCase
from io import StringIO

from .conf import (
    TEST_CENTER_NAME,
    LAB_VALUE_DICTS,
    LAB_VALUE_W_DIST_DICTS,
    TEST_PATIENT_LAB_SAMPLE_OUTPUT_PATH
)



class TestGeneratePatient(TestCase):
    def setUp(self):
        out = StringIO()
        call_command("load_gender_data", stdout=out)
        call_command("load_unit_data", stdout=out)
        call_command("load_distribution_data", stdout=out)
        call_command("load_lab_value_data", stdout=out)
        Center.objects.create(name=TEST_CENTER_NAME)

    def test_create_generic_patient_with_lab_sample(self):
        out = ""
        patient = Patient.create_generic(center=TEST_CENTER_NAME)
        sample = patient.create_lab_sample(
            sample_type="generic",
            date=None, # defaults to dt.now() if not provided
            save=True
        )

        for _dict in LAB_VALUE_DICTS:
            lab_value = PatientLabValue.create_lab_value_by_sample(
                sample=sample,
                **_dict
            )

            lab_value.set_unit_from_default()
            lab_value.set_norm_values_from_default()

        sample_2 = patient.create_lab_sample(
            sample_type="generic",
            date=None, # defaults to dt.now() if not provided
            save=True
        )

        lab_values = LabValue.objects.all()

        for lab_value in lab_values:
            try:
                patient_lab_value = PatientLabValue.create_lab_value_by_sample(
                    sample=sample_2,
                    lab_value_name = lab_value.name,
                )
                
                patient_lab_value.set_unit_from_default()
                patient_lab_value.set_norm_values_from_default()
                patient_lab_value.set_value_by_distribution()
            except Exception as e:
                out += f"\n{lab_value.name}: {e}\n"

        for value in sample.get_values():
            out += f"{value}\n"

        for value in sample_2.get_values():
            out += f"{value}\n"

        with open(TEST_PATIENT_LAB_SAMPLE_OUTPUT_PATH, "w") as f:
            f.write(out)

        return patient
    



from django.core.files import File
import io
from pathlib import Path
from typing import List


def prepare_bulk_frames(frame_paths: List[Path]):
    """
    Reads the frame paths into memory as Django File objects.
    This avoids 'seek of closed file' errors by using BytesIO for each frame.
    """
    for path in frame_paths:
        frame_number = int(path.stem.split("_")[1])
        with open(path, "rb") as f:
            content = f.read()
        file_obj = File(io.BytesIO(content), name=path.name)
        yield frame_number, file_obj



from django.db import models
from typing import List


class ExaminationManager(models.Manager):
    """
    Manager for Examination with custom query methods.
    """

    def get_by_natural_key(self, name: str) -> "Examination":
        return self.get(name=name)


class Examination(models.Model):
    """
    Represents an examination with associated details.

    Attributes:
        name (str): The unique name of the examination.
        name_de (str): The German name of the examination.
        name_en (str): The English name of the examination.
        examination_types (ManyToManyField): The types associated with the examination.
        date (DateField): The date of the examination.
        time (TimeField): The time of the examination.
    """

    name = models.CharField(max_length=100, unique=True)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)
    examination_types = models.ManyToManyField("ExaminationType", blank=True)

    objects = ExaminationManager()

    def __str__(self) -> str:
        """
        String representation of the examination.

        Returns:
            str: The name of the examination.
        """
        return self.name

    def natural_key(self) -> tuple:
        """
        Returns the natural key for the examination.

        Returns:
            tuple: The natural key consisting of the name.
        """
        return (self.name,)

    def get_available_findings(self) -> List["Finding"]:
        """
        Retrieves all findings associated with the examination.

        Returns:
            list: A list of findings related to the examination.
        """
        from endoreg_db.models import Finding

        findings: List[Finding] = [_ for _ in self.findings.all()]
        return findings

    class Meta:
        verbose_name = "Examination"
        verbose_name_plural = "Examinations"
        ordering = ["name"]



from .video import (
    get_videos_scheduled_for_frame_extraction, 
    extract_frames_from_video,
    extract_frames_from_videos,

    get_videos_scheduled_for_ocr,
    videos_scheduled_for_ocr_preflight,
    perform_ocr_on_video,
    perform_ocr_on_videos,

    videos_scheduled_for_initial_prediction_preflight,
    get_videos_scheduled_for_initial_prediction,
    get_multilabel_model,
    get_multilabel_classifier,
    get_crops,
    perform_initial_prediction_on_video,
    perform_initial_prediction_on_videos,

    videos_scheduled_for_prediction_import_preflight,
    get_videos_scheduled_for_prediction_import,
    import_predictions_for_video,
    import_predictions_for_videos,
    

    delete_frames_preflight,
    get_videos_scheduled_for_frame_deletion,
    delete_frames_for_video,
    delete_frames,
)

from .pdf import (
    get_pdf_files_scheduled_for_processing,
    process_pdf_file,
    process_pdf_files,
)


from django.db import models
from django.utils.translation import gettext_lazy as _

class PatientMedication(models.Model):
    patient = models.ForeignKey("Patient", on_delete= models.CASCADE)
    medication_indication = models.ForeignKey(
        "MedicationIndication", on_delete=models.CASCADE,
        related_name="patient_medications", null=True
    )

    medication = models.ForeignKey(
        'Medication', on_delete=models.CASCADE,
        blank=True,
        related_name='patient_medications'
    )

    intake_times = models.ManyToManyField(
        'MedicationIntakeTime', 
        related_name='patient_medications',
        blank=True,
    )

    unit = models.ForeignKey(
        'Unit', on_delete=models.CASCADE,
        null=True, blank=True
    )
    dosage = models.JSONField(
        null=True, blank=True
    )
    active = models.BooleanField(default=True)

    objects = models.Manager()

    class Meta:
        verbose_name = _('Patient Medication')
        verbose_name_plural = _('Patient Medications')

    @classmethod
    def create_by_patient_and_indication(cls, patient, medication_indication):
        from endoreg_db.models import MedicationIndication
        medication_indication: MedicationIndication
        patient_medication = cls.objects.create(patient=patient, medication_indication=medication_indication)
        patient_medication.save()
        
        return patient_medication


    def __str__(self):
        intake_times = self.intake_times.all()
        out = f"{self.medication} (Indication {self.medication_indication}) - "
        out += f"{self.dosage} - {self.unit} - "


        for intake_time in intake_times:
            out += f"{intake_time} - "

        return out
    
    


from django.db import models
from typing import List

# Serializer located in serializers/examination.py
from typing import Optional
from datetime import datetime


class PatientExamination(models.Model):
    patient = models.ForeignKey(
        "Patient", on_delete=models.CASCADE, related_name="patient_examinations"
    )
    examination = models.ForeignKey(
        "Examination", on_delete=models.CASCADE, null=True, blank=True
    )
    video = models.OneToOneField(
        "Video",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name="patient_examination",
    )
    date_start = models.DateField(null=True, blank=True)
    date_end = models.DateField(null=True, blank=True)
    hash = models.CharField(max_length=255, unique=True)

    # report_files
    class Meta:
        verbose_name = "Patient Examination"
        verbose_name_plural = "Patient Examinations"
        ordering = ["patient", "examination", "date_start"]

    @classmethod
    def get_or_create_pseudo_patient_examination_by_hash(
        cls,
        patient_hash: str,
        examination_hash: str,
        examination_name: Optional[str] = None,
    ):
        from endoreg_db.models import Patient, Examination

        created = False

        if PatientExamination.objects.filter(
            patient__patient_hash=patient_hash, hash=examination_hash
        ).exists():
            return PatientExamination.objects.get(
                patient__patient_hash=patient_hash, hash=examination_hash
            ), created

        patient, created = Patient.get_or_create_pseudo_patient_by_hash(patient_hash)
        if examination_name is not None:
            examination = Examination.objects.get(name=examination_name)
        else:
            examination = None

        patient_examination = cls.objects.create(
            patient=patient, examination=examination, hash=examination_hash
        )

        patient_examination.save()

        created = True
        return patient_examination, created

    def __str__(self):
        return f"{self.patient} - {self.examination} - {self.date_start}"

    # override save method to make sure that the hash is always set,
    # if none is existing generate an unique string

    def generate_default_hash(self):
        # create random hash
        import random
        import string

        _hash = "DEFAULT_HASH_" + "".join(
            random.choices(string.ascii_uppercase + string.digits, k=10)
        )

        return _hash

    def save(self, *args, **kwargs):
        if not self.hash:
            self.hash = self.generate_default_hash()
        super().save(*args, **kwargs)

    def get_patient_age_at_examination(self) -> int:
        """
        Returns the patient's age at the time of the examination.
        """
        from endoreg_db.models import Patient
        from datetime import datetime

        patient: Patient = self.patient
        dob = patient.get_dob()
        date_start = self.date_start
        return (date_start - dob).days // 365

    def get_available_findings(self):
        """
        Returns all findings that are associated with the examination of this patient examination.
        """
        from endoreg_db.models import Finding, Examination

        examination: Examination = self.examination
        findings: List[Finding] = [_ for _ in examination.get_available_findings()]
        return findings

    def get_findings(self):
        """
        Returns all findings that are associated with this patient examination.
        """
        from endoreg_db.models import PatientFinding

        patient_findings: List[PatientFinding] = [
            _ for _ in self.patient_findings.all()
        ]
        return patient_findings

    def get_indications(self):
        """
        Returns all indications that are associated with this patient examination.
        """
        from endoreg_db.models import PatientExaminationIndication

        indications: List[PatientExaminationIndication] = [
            _ for _ in self.indications.all()
        ]
        return indications

    def get_indication_choices(self):
        """
        Returns all indication choices that are associated with this patient examination.
        """
        from endoreg_db.models import ExaminationIndicationClassificationChoice

        choices: List[ExaminationIndicationClassificationChoice] = [
            _.indication_choice for _ in self.get_indications()
        ]
        return choices

    def create_finding(self, finding):
        """
        Adds a finding to this patient examination.
        """
        from endoreg_db.models import Finding, Examination, PatientFinding

        examination: Examination = self.examination
        assert examination

        finding: Finding

        patient_finding = PatientFinding.objects.create(
            patient_examination=self, finding=finding
        )

        patient_finding.save()

        return patient_finding

    def find_matching_video_from_patient(self):
        """
        Finds a video for this patient examination based on the patient's videos.
        For this, the videos date must be the same as the report file's date.
        #TODO add more criteria for matching: Examination type
        """
        videos = self.patient.video_set.filter(
            date=self.report_file.date, patient_examination__isnull=True
        )
        if videos:
            if len(videos) > 1:
                print(
                    f"Warning: Found more than one video for patient {self.patient} on date {self.report_file.date}. Choosing the first one."
                )
            return videos[0]
        else:
            videos = self.patient.video_set.filter(patient_examination__isnull=True)
            if len(videos) == 1:
                return videos[0]

        return None



# from endoreg_db.models import (
#     RawVideoFile,Center,VideoMeta
# )
    
# from datetime import datetime

# from django.core.management import call_command
# from django.test import TestCase
# from io import StringIO
# from .conf import (
#     TEST_VIDEO_DICT,
#     TEST_CENTER_NAME,
#     TEST_GASTROSCOPY_VIDEO_OUTPUT_PATH,
#     check_file_exists,
#     TEST_SALT
# )
# from endoreg_db.utils.hashs import (
#     get_hash_string
# )

# import json

# from pathlib import Path

# with open(TEST_GASTROSCOPY_VIDEO_OUTPUT_PATH, "w") as f:
#     f.write("Test Gastroscopy Video\n\n")

# class TestVideo(TestCase):
#     def setUp(self):
#         out = StringIO()
#         call_command("load_gender_data", stdout=out)
#         call_command("load_unit_data", stdout=out)
#         call_command("load_name_data", stdout=out)
#         call_command("load_center_data", stdout=out)
#         call_command("load_endoscope_data", stdout=out)
#         call_command("load_green_endoscopy_wuerzburg_data", stdout=out)


#         university_hospital_wuerzburg = Center.objects.get(name=TEST_CENTER_NAME)
#         assert university_hospital_wuerzburg, "Center not found"

#     def test_gastro_video(self):
#         from endoreg_db.models import SensitiveMeta
#         video_dict = TEST_VIDEO_DICT["gastroscopy"]
#         result_dict = video_dict["results"]
        
#         if check_file_exists(video_dict["path"]) and not video_dict["skip"]:
#             center = Center.objects.get(name=TEST_CENTER_NAME)
#             raw_video = RawVideoFile.create_from_file(
#                 file_path=video_dict["path"],
#                 center_name=TEST_CENTER_NAME,
#                 processor_name = video_dict["endoscopy_processor"],
#                 delete_source=False,
#             )

#             imported_video_path = raw_video.file.path
#             self.assertTrue(Path(imported_video_path).exists())
#             self.assertTrue(Path(imported_video_path).is_file())

#             video_meta = raw_video.video_meta
#             with open(TEST_GASTROSCOPY_VIDEO_OUTPUT_PATH, "a") as f:
#                 f.write(f"Video Meta:\n{video_meta}\n\n")
#                 f.write(f"FFMPEG Meta:\n{video_meta.ffmpeg_meta}\n\n")
#                 endo_roi_dict = video_meta.get_endo_roi()
#                 f.write("Endoscope ROI:\n")
#                 for key, value in endo_roi_dict.items():
#                     f.write(f"\t{key}:\t")
#                     f.write(f"{value}\n")
#                 f.write("\n")

#             extract_frames_result = raw_video.extract_frames(verbose=True)

#             with open(TEST_GASTROSCOPY_VIDEO_OUTPUT_PATH, "a") as f:
#                 f.write(f"Extract Frames Result:\n{extract_frames_result}\n\n")
                
#             # get frame paths:
#             frame_paths = raw_video.get_frame_paths()
#             for frame_path in frame_paths:
#                 self.assertTrue(Path(frame_path).exists())
#                 self.assertTrue(Path(frame_path).is_file())

#             # Extract Text info and append  to file:
#             raw_video.update_text_metadata(ocr_frame_fraction=0.001)
#             sensitive_meta:SensitiveMeta = raw_video.sensitive_meta
#             with open(TEST_GASTROSCOPY_VIDEO_OUTPUT_PATH, "a") as f:
#                 f.write(f"Text Info:\n{sensitive_meta}\n\n")

            
#             # check states (substitution, hashing, etc.)
#             # if state not true, then run functions to update states

#             # In Production, the sensitive metadata object must be verified by human annotation
#             sensitive_meta.state_verified = True
#             raw_video.save()

#             patient_hash = raw_video.sensitive_meta.get_patient_hash(salt=TEST_SALT)
#             patient_examination_hash = raw_video.sensitive_meta.get_patient_examination_hash(salt=TEST_SALT)
            
#             patient_hash_string = get_hash_string(
#                 first_name = sensitive_meta.patient_first_name,
#                 last_name = sensitive_meta.patient_last_name,
#                 dob=sensitive_meta.patient_dob,
#                 center_name = sensitive_meta.center.name,
#                 salt=TEST_SALT
#             )

#             patient_examination_hash_string = get_hash_string(
#                 first_name=sensitive_meta.patient_first_name,
#                 last_name=sensitive_meta.patient_last_name,
#                 dob=sensitive_meta.patient_dob,
#                 center_name=sensitive_meta.center.name,
#                 examination_date=sensitive_meta.examination_date,
#                 salt=TEST_SALT
#             )
            
#             with open(TEST_GASTROSCOPY_VIDEO_OUTPUT_PATH, "a") as f:
#                 f.write(f"Patient String for Hashing:\n\t{patient_hash_string}\n")
#                 f.write(f"Patient Hash:\n\t{patient_hash}\n\n")
#                 f.write(f"Patient Examination String for Hashing:\n\t{patient_examination_hash_string}\n")
#                 f.write(f"Patient Examination Hash:\n\t{patient_examination_hash}\n\n")

#             assert patient_hash==result_dict["patient_hash"], "Patient Hashes do not match"
#             assert patient_examination_hash==result_dict["patient_examination_hash"], "Patient Examination Hashes do not match"

#             frame_dir = raw_video.frame_dir
#             raw_video.delete_with_file()

#             self.assertFalse(Path(imported_video_path).exists())
#             self.assertFalse(Path(frame_dir).exists())

#             with open(TEST_GASTROSCOPY_VIDEO_OUTPUT_PATH, "a") as f:
#                 f.write("Video and Frames Deleted successfully\n\n")

#             return True

#     def test_load_video_legacy(self):
#         out = StringIO()
#         # import würzburg 1920x1080 anonymized videos
#         call_command("load_ai_model_data", stdout=out)






# ReportReaderConfig Class
# Description: This class is used to store the configuration of the ReportReader

# PATIENT_INFO_LINE_FLAG = "Patient: "
# ENDOSCOPE_INFO_LINE_FLAG = "Gerät: "
# EXAMINER_INFO_LINE_FLAG = "1. Unters.:"
# CUT_OFF_BELOW_LINE_FLAG = "________________"


# CUT_OFF_ABOVE_LINE_FLAGS = [
#     ENDOSCOPE_INFO_LINE_FLAG,
#     EXAMINER_INFO_LINE_FLAG,
# ]

# CUT_OFF_BELOW_LINE_FLAGS = [
#         CUT_OFF_BELOW_LINE_FLAG
#     ]

from django.db import models
from .report_reader_flag import ReportReaderFlag
from ..center import Center
from ..data_file import PdfType

class ReportReaderConfig(models.Model):

    locale = models.CharField(default="de_DE", max_length=10)
    first_names = models.ManyToManyField('FirstName', related_name='report_reader_configs')
    last_names = models.ManyToManyField('LastName', related_name='report_reader_configs')
    text_date_format = models.CharField(default = "%d.%m.%Y", max_length=10)
    patient_info_line_flag = models.ForeignKey(ReportReaderFlag, related_name='report_reader_configs_patient_info_line', on_delete=models.CASCADE)
    endoscope_info_line_flag = models.ForeignKey(ReportReaderFlag, related_name='report_reader_configs_endoscope_info_line', on_delete=models.CASCADE)
    examiner_info_line_flag = models.ForeignKey(ReportReaderFlag, related_name='report_reader_configs_examiner_info_line', on_delete=models.CASCADE)
    cut_off_below = models.ManyToManyField(ReportReaderFlag, related_name='report_reader_configs_cut_off_below')
    cut_off_above = models.ManyToManyField(ReportReaderFlag, related_name='report_reader_configs_cut_off_above')
    
    def __str__(self):
        return self.locale
    
    def update_names_by_center(self, center:Center, save = True):
        self.first_names.set(center.first_names.all())
        self.last_names.set(center.last_names.all())
        if save:
            self.save()

    def update_flags_by_pdf_type(self, pdf_type:PdfType, save = True):
        self.patient_info_line_flag = pdf_type.patient_info_line_flag
        self.endoscope_info_line_flag = pdf_type.endoscope_info_line_flag
        self.examiner_info_line_flag = pdf_type.examiner_info_line_flag
        self.cut_off_below.set(pdf_type.cut_off_below.all())
        self.cut_off_above.set(pdf_type.cut_off_above.all())
        if save:
            self.save()




from django import forms
from django.forms import ModelForm, Select
from endoreg_db.models.persons.patient.patient import Patient
from django import forms
from django.forms import ModelForm, Select
from endoreg_db.models.persons.patient.patient import Patient
from endoreg_db.models.persons.gender import Gender  
from endoreg_db.models.center import Center  
class PatientForm(ModelForm):
    gender = forms.ModelChoiceField(
        queryset=Gender.objects.all(),  
        empty_label="Select Gender",  
        widget=Select(attrs={'class': 'form-control'})
    )
    center = forms.ModelChoiceField(
        queryset=Center.objects.all(),  
        empty_label="Select Center",  
        widget=Select(attrs={'class': 'form-control'})
    )

    class Meta:
        model = Patient
        fields = '__all__'
        widgets = {
            'dob': forms.DateInput(attrs={'type': 'date', 'class': 'form-control'}),
        }



"""Module for Organ models."""
from django.db import models

class OrganManager(models.Manager):
    """Manager for Organ model."""

    def get_by_natural_key(self, name):
        """Retrieve an Organ by its natural key."""
        return self.get(name=name)
    
    def all_names(self):
        """Return a list of all organ names."""
        return list(self.all().values_list('name', flat=True))
    
class Organ(models.Model):
    """Model representing an organ."""
    
    name = models.CharField(max_length=100, unique=True)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    description_de = models.TextField(blank=True, null=True)
    description_en = models.TextField(blank=True, null=True)

    location_choices = models.ManyToManyField(
        'FindingLocationClassificationChoice', 
        blank=True, related_name='organs'
    )

    objects = OrganManager()
    
    def natural_key(self):
        """Return the natural key for the organ."""
        return (self.name,)
    
    def __str__(self):
        """Return string representation of the organ."""
        return str(self.name)


# models/data_file/import_classes/raw_pdf.py
# django db model "RawPdf"
# Class to store raw pdf file using django file field
# Class contains classmethod to create object from pdf file
# objects contains methods to extract text, extract metadata from text and anonymize text from pdf file uzing agl_report_reader.ReportReader class
# ------------------------------------------------------------------------------

from django.db import models

# from django.core.files.storage import FileSystemStorage
# from django.conf import settings
from django.core.exceptions import ValidationError
# from django.core.validators import FileExtensionValidator
# from endoreg_db.utils.file_operations import get_uuid_filename
# from icecream import ic

# from agl_report_reader.report_reader import ReportReader

from endoreg_db.utils.hashs import get_pdf_hash
# from ..metadata import SensitiveMeta

# setup logging to pdf_import.log
import logging

# import shutil
from pathlib import Path
from typing import TYPE_CHECKING

from ..base_classes.utils import (
    STORAGE_LOCATION,
)

if TYPE_CHECKING:
    from endoreg_db.models import SensitiveMeta

logger = logging.getLogger("pdf_import")

RAW_PDF_DIR_NAME = "raw_pdf"
RAW_PDF_DIR = STORAGE_LOCATION / RAW_PDF_DIR_NAME

if not RAW_PDF_DIR.exists():
    RAW_PDF_DIR.mkdir(parents=True)


class AbstractPdfFile(models.Model):
    pdf_hash = models.CharField(max_length=255, unique=True)
    pdf_type = models.ForeignKey(
        "PdfType",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    sensitive_meta = models.ForeignKey(
        "SensitiveMeta",
        on_delete=models.SET_NULL,
        related_name="report_files",
        null=True,
        blank=True,
    )
    center = models.ForeignKey(
        "Center",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    patient = models.ForeignKey(
        "Patient",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    examination = models.ForeignKey(
        "PatientExamination",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    examiner = models.ForeignKey(
        "Examiner",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    text = models.TextField(blank=True, null=True)

    created_at = models.DateTimeField(auto_now_add=True)

    if TYPE_CHECKING:
        sensitive_meta: "SensitiveMeta"

    class Meta:
        abstract = True

    def __str__(self):
        str_repr = f"{self.pdf_hash} ({self.pdf_type}, {self.center})"
        return str_repr

    def delete_with_file(self):
        file_path = Path(self.file.path)
        if file_path.exists():
            file_path.unlink()
            logger.info("File removed: %s", file_path)

        r = self.delete()
        return r

    def update(self, save=True, verbose=True):
        try:
            self.text, self.anonymized_text, self.raw_meta = self.process_file(
                verbose=verbose
            )
            self.state_report_processed = True
            self.state_report_processing_required = False

            if save:
                self.save()

            return True

        except (IOError, ValueError) as e:
            logger.error("Error processing file: %s", self.file.path)
            logger.error(e)
            return False

    # override save method to get patient and examination from sensitive meta
    def save(self, *args, **kwargs):
        if not self.patient and self.sensitive_meta:
            self.patient = self.sensitive_meta.pseudo_patient
            self.examination = self.sensitive_meta.pseudo_examination
            self.center = self.sensitive_meta.center

        super().save(*args, **kwargs)



from endoreg_db.models import (
    Patient, 
    Center, Event,
    PatientLabValue, PatientFinding, FindingMorphologyClassification,
    Finding,
    FindingLocationClassification,
    PatientMedication, PatientMedicationSchedule, MedicationIndication,
    MedicationIndicationType, MedicationSchedule
)
from datetime import datetime

from django.core.management import call_command
from django.test import TestCase
from io import StringIO
from .conf import (
    TEST_CENTER_NAME,
    TEST_PATIENT_MEDICATION_OUTPUT_PATH,
    TEST_ANTICOAGULATION_INDICATION_TYPES
)

# Note
# A MedicationIndication can have multiple MedicationSchedules 
# (e.g., pulmonary embolism medication starts with high dose for 10d and is then reduced)

# -> A MedicationSchedule defines the dosage, intake_times, and duration for ONE DRUG

# A PatientMedication has a patient as foreign key and may point to one MedicationSchedule

# A PatientMedicationSchedule has a patient as foreign key and may point to multiple PatientMedications


# PatientMedicationSchedule may point to multiple PatientMedications

#### Example
# - model: endoreg_db.medication_indication
#   fields:
#     name: "te_prevention-after_hip-apixaban"
#     events: ["hip_replacement"]
#     medication_schedules: [
#       "apixaban-2.5mg-twice_daily_32d", # min 32d, max 38d
#     ]
#     indication_type: "thromboembolism-prevention-after_hip_replacement"
#     sources: [
#       "fachinformation-apixaban-eliquis_2.5mg-bristol_meyer_squibbs"
#     ]

# Workflow to create a PatientMedicationSchedule for a single MedicationIndication:
# get medication_indication object and use classmethod create_by_patient_and_indication


class TestGeneratePatientMedication(TestCase):
    def setUp(self):
        out = StringIO()
        call_command("load_gender_data", stdout=out)
        call_command("load_unit_data", stdout=out)
        call_command("load_disease_data", stdout=out)
        call_command("load_event_data", stdout=out)
        call_command("load_medication_data", stdout=out)
        Center.objects.create(name=TEST_CENTER_NAME)

    def test_patient_anticoagulation(self):
        patient = Patient.create_generic(center=TEST_CENTER_NAME)
        out = ""

        pms = PatientMedicationSchedule.objects.create(
            patient = patient,
        )


        for indication_type in TEST_ANTICOAGULATION_INDICATION_TYPES:
            out += f"IndicationType - {indication_type}:\n" 
            indication_type = MedicationIndicationType.objects.get(name=indication_type)
            medication_indications = MedicationIndication.objects.filter(
                indication_type=indication_type
            )
        

            for medication_indication in medication_indications:
                medication_schedules = medication_indication.medication_schedules.all()
                out += f"\tMedicationIndication - {medication_indication}:\n"

                for medication_schedule in medication_schedules:
                    out += f"\t\tMedicationSchedule - {medication_schedule}:\n"
                    _patient_medication = pms.create_patient_medication_from_medication_schedule(
                        medication_schedule=medication_schedule,
                        medication_indication = medication_indication,
                        start_date=datetime.now()
                    )

                    out += f"\t\t\tPatientMedication - {_patient_medication}\n"

        pms = PatientMedicationSchedule.objects.get(pk = pms.pk)
        out += f"\n---\nResulting PatientMedicationSchedule:\n{pms}\n"
        out += "\n---\nPatientMedications:\n"
        for pm in pms.get_patient_medication():
            out += f"\t{pm}\n"
                


        with open(TEST_PATIENT_MEDICATION_OUTPUT_PATH, "w") as f:
            _ = out
            f.write(_)



'''Model for medication indication type.'''
from django.db import models

class MedicationIndicationTypeManager(models.Manager):
    '''Manager for the medication indication type model.'''
    def get_by_natural_key(self, name):
        '''Retrieve a medication indication type by its natural key.'''
        return self.get(name=name)

class MedicationIndicationType(models.Model):
    '''Model representing a medication indication type.'''
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)

    objects = MedicationIndicationTypeManager()

    def natural_key(self):
        '''Return the natural key for the medication indication type.'''
        return (self.name,)

    def __str__(self):
        return str(self.name)
    
    @classmethod
    def get_random_indication_by_type(cls, name) -> "MedicationIndication":
        '''Return a random medication indication of the given type.'''
        return cls.objects.get(name=name).medication_indications.order_by('?').first()
    

    def get_random_medication_indication(self):
        '''Return a random medication indication of this type.'''
        from endoreg_db.models import MedicationIndication
        return MedicationIndication.objects.filter(indication_type=self).order_by('?').first()



from rest_framework import serializers
from endoreg_db.models import (
    ImageClassificationPrediction,
)


class ImageClassificationPredictionSerializer(serializers.ModelSerializer):
    class Meta:
        model = ImageClassificationPrediction
        fields = "__all__"



from .inference_dataset import InferenceDataset
from torch.utils.data import DataLoader
import torch
from torch import nn
import json
from .postprocess import (
    concat_pred_dicts,
    make_smooth_preds,
    find_true_pred_sequences
)
import numpy as np

sample_config = {
    # mean and std for normalization
    "mean": (0.45211223, 0.27139644, 0.19264949),
    "std": (0.31418097, 0.21088019, 0.16059452),
    # Image Size
    "size_x": 716,
    "size_y": 716,
    # how to wrangle axes of the image before putting them in the network
    "axes": [2,0,1],  # 2,1,0 for opencv
    "batchsize": 16,
    "num_workers": 0, # always 1 for Windows systems # FIXME: fix celery crash if multiprocessing
    # maybe add sigmoid after prediction?
    "activation": nn.Sigmoid(),
    "labels": [
        'appendix',  'blood',  'diverticule',  'grasper',  'ileocaecalvalve',  'ileum',  'low_quality',  'nbi',  'needle',  'outside',  'polyp',  'snare',  'water_jet',  'wound'
    ]
}

class Classifier():
    def __init__(self, model=None, config=sample_config, verbose = False):
        self.config = config
        self.model = model
        self.verbose = verbose

    def pipe(self, paths, crops, verbose = None):
        if verbose is None:
            verbose = self.verbose

        dataset = InferenceDataset(paths, crops, self.config)
        if verbose:
            print("Dataset created")

        dl = DataLoader(
            dataset=dataset,
            batch_size=self.config["batchsize"],
            num_workers=self.config["num_workers"],
            shuffle = False,
            pin_memory=True
        )
        if verbose:
            print("Dataloader created")

        predictions = []

        with torch.inference_mode():
            if self.verbose:
                print("Starting inference")
            for i,batch in enumerate(dl):
                prediction = self.model(batch.cuda())
                prediction = self.config["activation"](prediction).cpu().tolist()#.numpy().tolist()
                predictions += prediction
                if self.verbose and i==0:
                    print("First batch done")

        return predictions

    def __call__(self, image, crop=None):
        return self.pipe([image], [crop])

    def readable(self, predictions):
        return {label: prediction for label, prediction in zip(self.config["labels"], predictions)}
    
    def get_prediction_dict(self, predictions, paths):
        json_dict = {
            "labels": self.config["labels"],
            "paths": paths,
            "predictions": predictions
        }
        
        return json_dict
    
    def get_prediction_json(self, predictions, paths, json_target_path: str = None):
        if not json_target_path:
            json_target_path = "predictions.json"

        json_dict = self.get_prediction_dict(predictions, paths)

        with open(json_target_path, 'w') as f:
            json.dump(json_dict, f)
        
        if self.verbose:
            print(f"Saved predictions to {json_target_path}")

    def post_process_predictions(self, pred_dicts, window_size_s=1, fps=50, min_seq_len_s = 0.5):
        '''
        pred_dicts: list of dictionaries with the same keys
        window_size_s: size of the window in seconds for smoothing
        fps: frames per second
        min_seq_len_s: minimum length of a sequence in seconds

        Returns:
        predictions: concatenated predictions
        smooth_predictions: smoothed predictions
        binary_predictions: binary predictions
        raw_sequences: raw sequences
        filtered_sequences: filtered sequences
        '''
        # Concatenate the predictions
        predictions = concat_pred_dicts(pred_dicts)

        smooth_predictions = {key:[] for key in predictions.keys()}
        for key in predictions.keys():
            smooth_predictions[key] = make_smooth_preds(
                predictions[key],
                window_size_s=window_size_s,
                fps=fps
            )

        binary_predictions = {}
        for key in smooth_predictions.keys():
            binary_predictions[key] = np.array([p > 0.5 for p in smooth_predictions[key]])

        raw_sequences = {}
        for key in binary_predictions.keys():
            raw_sequences[key] = find_true_pred_sequences(binary_predictions[key])

        filtered_sequences = {}
        min_seq_len = int(min_seq_len_s * fps)
        for key in raw_sequences.keys():
            filtered_sequences[key] = [s for s in raw_sequences[key] if s[1] - s[0] > min_seq_len]

        return predictions, smooth_predictions, binary_predictions, raw_sequences, filtered_sequences

    def post_process_predictions_serializable(
        self, pred_dicts,
        window_size_s = 1, fps = 50,
        min_seq_len_s = 0.5
    ):
        result = self.post_process_predictions(
            pred_dicts,
            window_size_s,
            fps, min_seq_len_s
        )

        for i, _dict in enumerate(result):
            _keys = list(_dict.keys())
            for key in _keys:
                # if numpy array
                if hasattr(_dict[key], "tolist"):
                    result[i][key] = _dict[key].tolist()
                
                # check if list of tuples
                # if so, make sure each tuple has 2 elements and split to two lists (start, stop)
                if all(isinstance(x, tuple) for x in _dict[key]):
                    if all(len(x) == 2 for x in _dict[key]):
                        result[i][f"{key}_start"] = [int(x[0]) for x in _dict[key]]
                        result[i][f"{key}_stop"] = [int(x[1]) for x in _dict[key]]
                        del result[i][key]
        
        # make dict of dicts
        result_dict = {
            "predictions": result[0],
            "smooth_predictions": result[1],
            "binary_predictions": result[2],
            "raw_sequences": result[3],
            "filtered_sequences": result[4]
        }

        return result_dict
                


# Generated by Django 5.1.6 on 2025-03-13 14:15

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('endoreg_db', '0007_rename_pseudo_patient_video_patient_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='reportfile',
            name='patient_examination',
        ),
        migrations.AddField(
            model_name='rawpdffile',
            name='examination',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='endoreg_db.patientexamination'),
        ),
        migrations.AddField(
            model_name='rawpdffile',
            name='examiner',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='endoreg_db.examiner'),
        ),
        migrations.AddField(
            model_name='rawvideofile',
            name='examination',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='endoreg_db.patientexamination'),
        ),
        migrations.AddField(
            model_name='reportfile',
            name='examination',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='endoreg_db.patientexamination'),
        ),
        migrations.AlterField(
            model_name='reportfile',
            name='examiner',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='endoreg_db.examiner'),
        ),
        migrations.AlterField(
            model_name='reportfile',
            name='patient',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='endoreg_db.patient'),
        ),
    ]



from django.db import models

class ProductGroupManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class ProductGroup(models.Model):
    objects = ProductGroupManager()

    name = models.CharField(max_length=255)
    name_de = models.CharField(max_length=255, null=True)
    name_en = models.CharField(max_length=255, null=True)
    
    def natural_key(self):
        return (self.name,)
    
    def __str__(self):
        return self.name
    
    def get_products(self):
        from endoreg_db.models import Product
        return Product.objects.filter(product_group=self)
    
    def get_reference_product(self):
        from endoreg_db.models import ReferenceProduct
        return ReferenceProduct.objects.get(product_group=self)
    



import subprocess
from django.db import models

# Django db class to store network devices (e.g., servers, clients, switches, etc.)

class NetworkDeviceManager(models.Manager):
    """Custom manager for NetworkDevice that defines name as natural key."""
    def get_by_natural_key(self, name):
        """Return the network device with the given name as its natural key."""
        return self.get(name=name)
    
class NetworkDevice(models.Model):
    """Django model representing a network device."""
    name = models.CharField(max_length=255)
    ip = models.GenericIPAddressField(blank=True, null=True)
    description = models.CharField(max_length=255)
    device_type = models.ForeignKey("NetworkDeviceType", on_delete=models.CASCADE)
    online = models.BooleanField(default=False)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    objects = NetworkDeviceManager()

    def __str__(self):
        """Return the device name."""
        return str(self.name)

    # pylint: disable=too-few-public-methods
    class Meta:
        """Meta options for the NetworkDevice model."""
        db_table = 'network_devices'
        ordering = ['name']
        
    def natural_key(self):
        """Return a tuple representing the natural key for this network device."""
        return (self.name,)
    
    def ping(self):
        """Check device availability by sending one ping request."""
        target_ip = self.ip

        # Define the command
        command = ['ping', '-c', '1', target_ip]

        # Run the command
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        # Wait for the process to terminate
        process.wait()

        # Check the return code
        return_code = process.returncode
        
        # Return True if the return code is 0, False otherwise
        self.online = return_code == 0
        self.save()
        return self.online





from django.http import JsonResponse
from django.middleware.csrf import get_token

# New view to return the CSRF token in JSON format
def csrf_token_view(request):
    token = get_token(request)
    return JsonResponse({'csrf_token': token})


from django.db import models

class PatientFindingLocation(models.Model):
    location_classification = models.ForeignKey('FindingLocationClassification', on_delete=models.CASCADE, related_name='patient_finding_locations')
    location_choice = models.ForeignKey('FindingLocationClassificationChoice', on_delete=models.CASCADE, related_name='patient_finding_locations')
    subcategories = models.JSONField(blank=True, null=True)
    numerical_descriptors = models.JSONField(blank=True, null=True)

    class Meta:
        verbose_name = 'Patient Finding Location'
        verbose_name_plural = 'Patient Finding Locations'
        ordering = ['location_classification', 'location_choice']

    def __str__(self):
        return f"{self.location_classification} - {self.location_choice}"
    
    # override save method to do the following:
    # - check if location_choice is in location_classification.choices
    # - check if subcategories and numerical_descriptors exist
    # - if not set, fetch them from location_choice and set them

    def save(self, *args, **kwargs):
        if self.location_choice not in self.location_classification.choices.all():
            raise ValueError("location_choice must be in location_classification.choices")

        if not self.subcategories:
            self.subcategories = self.location_choice.subcategories

        if not self.numerical_descriptors:
            self.numerical_descriptors = self.location_choice.numerical_descriptors

        super().save(*args, **kwargs)

    def set_subcategory(self, subcategory_name, subcategory_value):
        """
        Sets a subcategory for this location.
        """
        assert subcategory_name in self.subcategories, "Subcategory must be in subcategories."
        self.subcategories[subcategory_name]["value"] = subcategory_value
        self.save()

        return self.subcategories[subcategory_name]

    def set_random_subcategories(self):
        """
        Sets random subcategories for this location if they are required.
        """
        import random
        if not self.subcategories or not self.numerical_descriptors:
            self.save()

        self.refresh_from_db()
        assert self.subcategories, "Subcategories must be set."

        subcategories = self.subcategories
        # print("SUBCATS")
        # print(subcategories)

        # subcategories is dict with keys as subcategory names and values as dict with keys as "choices" (List of str) and "required" (bool)
        # for each subcategory, set a random choice if it is required
        for subcategory_name, subcategory_dict in subcategories.items():
            if subcategory_dict["required"]:
                subcategory_choice = random.choice(subcategory_dict["choices"])
                self.subcategories[subcategory_name]["value"] = subcategory_choice

        self.save()

        return self.subcategories
    
    def set_random_numerical_descriptor(self, descriptor_name):
        """
        Sets a random numerical descriptor for this location.
        """
        import random
        if descriptor_name not in self.numerical_descriptors:
            raise ValueError("Descriptor name must be in numerical descriptors.")

        numerical_descriptor = self.numerical_descriptors[descriptor_name]
        min_value = numerical_descriptor["min"]
        max_value = numerical_descriptor["max"]

        assert min_value <= max_value, "Min value must be less than or equal to max value."

        random_value = random.uniform(min_value, max_value)
        self.numerical_descriptors[descriptor_name]["value"] = random_value
        self.save()

        return self.numerical_descriptors[descriptor_name]


    def set_random_numerical_descriptors(self):
        """
        Sets random numerical descriptors for this location if they are required.
        """
        import random
        from endoreg_db.models import FindingLocationClassificationChoice
        if not self.subcategories or not self.numerical_descriptors:
            self.save()

        numerical_descriptors = self.numerical_descriptors
        numerical_descriptor = {}

        # numerical_descriptors is dict with keys as numerical descriptor names 
        # and values as dict with keys as "min" (float), "max" (float), required (bool), distribution_name (str)
        # distribution name can be either "uniform" or "normal"
        # for each numerical descriptor, set a random value between min and max
        for numerical_descriptor_name, numerical_descriptor_dict in numerical_descriptors.items():
            min_value = numerical_descriptor_dict["min"]
            max_value = numerical_descriptor_dict["max"]

            assert min_value <= max_value, "Min value must be less than or equal to max value."


            random_value = random.uniform(min_value, max_value)
            numerical_descriptor[numerical_descriptor_name] = random_value

        self.numerical_descriptors = numerical_descriptor
        self.save()

        return numerical_descriptor



# Generated by Django 5.1.6 on 2025-03-13 08:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('endoreg_db', '0005_rawvideofile_frame_count_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='frame',
            name='extracted',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='rawframe',
            name='extracted',
            field=models.BooleanField(default=False),
        ),
    ]



# models/data_file/import_classes/raw_pdf.py
# django db model "RawPdf"
# Class to store raw pdf file using django file field
# Class contains classmethod to create object from pdf file
# objects contains methods to extract text, extract metadata from text and anonymize text from pdf file uzing agl_report_reader.ReportReader class
# ------------------------------------------------------------------------------

from django.db import models
from django.core.files.storage import FileSystemStorage
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import FileExtensionValidator
from endoreg_db.utils.file_operations import get_uuid_filename
from icecream import ic

from agl_report_reader.report_reader import ReportReader

from endoreg_db.utils.hashs import get_pdf_hash
from ..metadata import SensitiveMeta
from ..base_classes.abstract_pdf import AbstractPdfFile

# setup logging to pdf_import.log
import logging

import shutil
from pathlib import Path

from ..base_classes.utils import (
    STORAGE_LOCATION,
)

logger = logging.getLogger("pdf_import")

RAW_PDF_DIR_NAME = "raw_pdf"
RAW_PDF_DIR = STORAGE_LOCATION / RAW_PDF_DIR_NAME

if not RAW_PDF_DIR.exists():
    RAW_PDF_DIR.mkdir(parents=True)


class RawPdfFile(AbstractPdfFile):
    file = models.FileField(
        upload_to=f"{RAW_PDF_DIR_NAME}/",
        validators=[FileExtensionValidator(allowed_extensions=["pdf"])],
        storage=FileSystemStorage(location=STORAGE_LOCATION.resolve().as_posix()),
    )

    patient = models.ForeignKey(
        "Patient",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="raw_pdf_files",
    )

    state_report_processing_required = models.BooleanField(default=True)
    state_report_processed = models.BooleanField(default=False)
    raw_meta = models.JSONField(blank=True, null=True)
    # report_file = models.OneToOneField("ReportFile", on_delete=models.CASCADE, null=True, blank=True)
    sensitive_meta = models.ForeignKey(
        "SensitiveMeta",
        on_delete=models.SET_NULL,
        related_name="raw_pdf_files",
        null=True,
        blank=True,
    )

    report_file = models.ForeignKey(
        "ReportFile",
        on_delete=models.SET_NULL,
        related_name="raw_pdf_files",
        null=True,
        blank=True,
    )

    anonymized_text = models.TextField(blank=True, null=True)

    @classmethod
    def create_from_file(
        cls,
        file_path: Path,
        center_name,
        save=True,
        delete_source=True,
    ):
        from endoreg_db.models import Center

        logger.info(f"Creating RawPdfFile object from file: {file_path}")
        ic(f"Creating RawPdfFile object from file: {file_path}")

        new_file_name, uuid = get_uuid_filename(file_path)

        pdf_hash = get_pdf_hash(file_path)
        ic(pdf_hash)
        new_file_path = RAW_PDF_DIR / new_file_name
        # check if pdf file already exists

        if cls.objects.filter(pdf_hash=pdf_hash).exists():
            existing_pdf_file = cls.objects.filter(pdf_hash=pdf_hash).get()
            logger.warning(f"RawPdfFile with hash {pdf_hash} already exists")
            ic(f"RawPdfFile with hash {pdf_hash} already exists")

            existing_pdf_file.verify_existing_file(fallback_file=file_path)

            return existing_pdf_file

        else:
            ic(f"No existing pdf file found for hash {pdf_hash}")

        # assert pdf_type_name is not None, "pdf_type_name is required"
        assert center_name is not None, "center_name is required"

        # pdf_type = PdfType.objects.get(name=pdf_type_name)
        center = Center.objects.get(name=center_name)

        logger.info(f"Copying file to {new_file_path}")
        ic(f"Copying file to {new_file_path}")
        _success = shutil.copy(file_path, new_file_path)

        # validate copy operation by comparing hashs
        assert get_pdf_hash(new_file_path) == pdf_hash, "Copy operation failed"

        raw_pdf = cls(
            file=new_file_path.resolve().as_posix(),
            pdf_hash=pdf_hash,
            # pdf_type=pdf_type,
            center=center,
        )
        raw_pdf.save()
        logger.info(f"RawPdfFile object created: {raw_pdf}")
        ic(f"RawPdfFile object created: {raw_pdf}")

        # remove source file
        if delete_source:
            file_path.unlink()
            logger.info(f"Source file removed: {file_path}")
            ic(f"Source file removed: {file_path}")

        if save:
            raw_pdf.save()

        return raw_pdf

    def save(self, *args, **kwargs):
        if not self.file.name.endswith(".pdf"):
            raise ValidationError("Only PDF files are allowed")

        if not self.pdf_hash:
            self.pdf_hash = get_pdf_hash(self.file.path)

        super().save(*args, **kwargs)

    def verify_existing_file(self, fallback_file):
        if not Path(self.file.path).exists():
            logger.warning(f"File not found: {self.file.path}")
            logger.warning(f"Using fallback file: {fallback_file}")
            ic(f"File not found: {self.file.path}")
            ic(f"Copy fallback file: {fallback_file} to existing filepath")

            shutil.copy(fallback_file, self.file.path)

            self.save()

    def process_file(self, verbose=False):
        pdf_path = self.file.path
        rr_config = self.get_report_reader_config()

        rr = ReportReader(
            **rr_config
        )  # FIXME In future we need to pass a configuration file
        # This configuration file should be associated with pdf type

        text, anonymized_text, report_meta = rr.process_report(
            pdf_path, verbose=verbose
        )

        self.text = text
        self.anonymized_text = anonymized_text

        report_meta["center_name"] = self.center.name
        if not self.sensitive_meta:
            sensitive_meta = SensitiveMeta.create_from_dict(report_meta)
            self.sensitive_meta = sensitive_meta

        else:
            # update existing sensitive meta
            sensitive_meta = self.sensitive_meta
            sensitive_meta.update_from_dict(report_meta)

        self.raw_meta = report_meta

        sensitive_meta.save()
        self.save()

        return text, anonymized_text, report_meta

    def get_report_reader_config(self):
        from endoreg_db.models import PdfType, Center
        from warnings import warn

        if not self.pdf_type:
            warn("PdfType not set, using default settings")
            pdf_type = PdfType.default_pdf_type()
        else:
            pdf_type: PdfType = self.pdf_type
        center: Center = self.center
        if pdf_type.endoscope_info_line:
            endoscope_info_line = pdf_type.endoscope_info_line.value

        else:
            endoscope_info_line = None
        settings_dict = {
            "locale": "de_DE",
            "employee_first_names": [_.name for _ in center.first_names.all()],
            "employee_last_names": [_.name for _ in center.last_names.all()],
            "text_date_format": "%d.%m.%Y",
            "flags": {
                "patient_info_line": pdf_type.patient_info_line.value,
                "endoscope_info_line": endoscope_info_line,
                "examiner_info_line": pdf_type.examiner_info_line.value,
                "cut_off_below": [_.value for _ in pdf_type.cut_off_below_lines.all()],
                "cut_off_above": [_.value for _ in pdf_type.cut_off_above_lines.all()],
            },
        }

        return settings_dict

    def get_or_create_report_file(self):
        from endoreg_db.models import ReportFile

        if self.report_file:
            report_file = self.report_file

        elif ReportFile.objects.filter(pdf_hash=self.pdf_hash).exists():
            report_file = ReportFile.objects.filter(pdf_hash=self.pdf_hash).get()
            self.report_file = report_file
            self.save()
        else:
            # TODO  Make sure all required states are set
            patient = self.sensitive_meta.get_or_create_pseudo_patient()
            examiner = self.sensitive_meta.get_or_create_pseudo_examiner()
            patient_examination = (
                self.sensitive_meta.get_or_create_pseudo_patient_examination()
            )

            report_file = ReportFile.objects.create(
                pdf_hash=self.pdf_hash,
                center=self.center,
                sensitive_meta=self.sensitive_meta,
                patient=patient,
                examiner=examiner,
                examination=patient_examination,
                text=self.anonymized_text,
            )

            report_file.save()
            self.report_file = report_file
            self.save()

        return report_file



# Utility
from .unit import Unit
from .information_source import InformationSource
from .emission import EmissionFactor

# Center
from .center import (
    Center,
    CenterProduct,
    CenterResource,
    CenterWaste,
)

# Persons
from .persons import (
    Gender,
    Person,
    Patient,
    PatientForm,
    PatientEvent,
    PatientDisease,
    PatientLabSample,
    PatientLabSampleType,
    PatientLabValue,
    PatientMedication,
    PatientMedicationSchedule,
    PatientExaminationIndication,
    Examiner,
    ExaminerSerializer,
    PortalUserInfo,
    Profession,
    FirstName,
    LastName,
)

# Network
from .network import (
    NetworkDevice,
    NetworkDeviceType,
    AglService,
)

# Logging
from .logging import (
    AbstractLogEntry,
    NetworkDeviceLogEntry,
    LogType,
    AglServiceLogEntry,
)

# Patient
from .patient import (
    PatientExamination,
    PatientFinding,
    PatientFindingLocation,
    PatientFindingMorphology,
    PatientFindingIntervention,
)

# Organ
from .organ import Organ

# LX
from .lx import (
    LxClientType,
    LxClientTag,
    LxClient,
    LxIdentity,
    LxIdentityType,
    LxPermission,
    LxUser,
)

# Contraindication
from .contraindication import Contraindication

# Finding
from .finding import (
    Finding,
    FindingType,
    FindingLocationClassification,
    FindingLocationClassificationChoice,
    FindingMorphologyClassificationType,
    FindingMorphologyClassificationChoice,
    FindingMorphologyClassification,
    FindingIntervention,
    FindingInterventionType,
)

# Case template
from .case_template import (
    CaseTemplate,
    CaseTemplateType,
    CaseTemplateRule,
    CaseTemplateRuleType,
    CaseTemplateRuleValue,
    CaseTemplateRuleValueType,
)

# Rules
from .rules import (
    Rule,
    RuleType,
    Ruleset,
    RuleApplicator,
)

# Disease
from .disease import Disease, DiseaseClassification, DiseaseClassificationChoice

# Event
from .event import Event, EventClassification, EventClassificationChoice

# Laboratory
from .laboratory import LabValue

# Medication
from .medication import (
    Medication,
    MedicationSchedule,
    MedicationIntakeTime,
    MedicationIndication,
    MedicationIndicationType,
)

# Examination
from .examination import (
    Examination,
    ExaminationType,
    ExaminationTime,
    ExaminationTimeType,
    ExaminationIndication,
    ExaminationIndicationClassification,
    ExaminationIndicationClassificationChoice,
)

# Report Reader
from .report_reader import (
    ReportReaderConfig,
    ReportReaderFlag,
)

# Data file
from .data_file import (
    Frame,
    ReportFile,
    Video,
    LabelVideoSegment,
    LabelRawVideoSegment,
    SensitiveMeta,
    PdfMeta,
    PdfType,
    VideoMeta,
    RawFrame,
    FFMpegMeta,
    VideoImportMeta,
    RawPdfFile,
    RawVideoFile,
)

# Label
from .label import Label, LabelType, LabelSet

# Annotation
from .annotation import (
    ImageClassificationAnnotation,
    BinaryClassificationAnnotationTask,
    AnonymousImageAnnotation,
    DroppedName,
    AnonymizedImageLabel,
    AnonymizedFile,
    UploadedFile,
    VideoSegmentationLabel,
    VideoSegmentationAnnotation,
    VideoSegmentationLabelSet,
)

# Other
from .other import (
    Material,
    Resource,
    TransportRoute,
    Waste,
    BaseValueDistribution,
    NumericValueDistribution,
    SingleCategoricalValueDistribution,
    MultipleCategoricalValueDistribution,
    DateValueDistribution,
)

# Product
from .product import (
    Product,
    ProductMaterial,
    ProductGroup,
    ReferenceProduct,
    ProductWeight,
)

# AI models
from .ai_model import (
    ModelMeta,
    ModelType,
    ActiveModel,
    AiModel,
)

# Quiz
from .quiz import (
    QuizAnswer,
    QuizAnswerType,
    QuizQuestion,
    QuizQuestionType,
)

# Prediction
from .prediction import (
    ImageClassificationPrediction,
    VideoPredictionMeta,
    RawVideoPredictionMeta,
)

# Hardware
from .hardware import (
    EndoscopyProcessor,
    Endoscope,
    EndoscopeType,
)

# Questionnaires
from .questionnaires import TtoQuestionnaire

__all__ = [
    "ReportReaderConfig",
    "ReportReaderFlag",
    "Unit",
    "InformationSource",
    "EmissionFactor",
    "Center",
    "CenterProduct",
    "CenterResource",
    "CenterWaste",
    "RawVideoPredictionMeta",
    "Gender",
    "Person",
    "Patient",
    "PatientForm",
    "PatientEvent",
    "PatientDisease",
    "PatientLabSample",
    "PatientLabSampleType",
    "PatientLabValue",
    "PatientMedication",
    "PatientMedicationSchedule",
    "PatientExaminationIndication",
    "Examiner",
    "ExaminerSerializer",
    "PortalUserInfo",
    "RawFrame",
    "Profession",
    "FirstName",
    "LastName",
    "NetworkDevice",
    "NetworkDeviceType",
    "AglService",
    "AbstractLogEntry",
    "NetworkDeviceLogEntry",
    "LogType",
    "AglServiceLogEntry",
    "PatientExamination",
    "PatientFinding",
    "PatientFindingLocation",
    "PatientFindingMorphology",
    "PatientFindingIntervention",
    "Organ",
    "LxClientType",
    "LxClientTag",
    "LxClient",
    "LxIdentity",
    "LxIdentityType",
    "LxPermission",
    "LxUser",
    "Contraindication",
    "Finding",
    "FindingType",
    "FindingLocationClassification",
    "FindingLocationClassificationChoice",
    "FindingMorphologyClassificationType",
    "FindingMorphologyClassificationChoice",
    "FindingMorphologyClassification",
    "FindingIntervention",
    "FindingInterventionType",
    "CaseTemplate",
    "CaseTemplateType",
    "CaseTemplateRule",
    "CaseTemplateRuleType",
    "CaseTemplateRuleValue",
    "CaseTemplateRuleValueType",
    "Rule",
    "RuleType",
    "Ruleset",
    "RuleApplicator",
    "Disease",
    "DiseaseClassification",
    "DiseaseClassificationChoice",
    "Event",
    "EventClassification",
    "EventClassificationChoice",
    "LabValue",
    "Medication",
    "MedicationSchedule",
    "MedicationIntakeTime",
    "MedicationIndication",
    "MedicationIndicationType",
    "Examination",
    "ExaminationType",
    "ExaminationTime",
    "ExaminationTimeType",
    "ExaminationIndication",
    "ExaminationIndicationClassification",
    "ExaminationIndicationClassificationChoice",
    "Label",
    "LabelType",
    "LabelSet",
    "LabelRawVideoSegment",
    "ImageClassificationAnnotation",
    "BinaryClassificationAnnotationTask",
    "AnonymousImageAnnotation",
    "DroppedName",
    "AnonymizedImageLabel",
    "AnonymizedFile",
    "UploadedFile",
    "VideoSegmentationLabel",
    "VideoSegmentationAnnotation",
    "Material",
    "Resource",
    "TransportRoute",
    "Waste",
    "BaseValueDistribution",
    "NumericValueDistribution",
    "SingleCategoricalValueDistribution",
    "MultipleCategoricalValueDistribution",
    "DateValueDistribution",
    "Product",
    "ProductMaterial",
    "ProductGroup",
    "ReferenceProduct",
    "ProductWeight",
    "ModelMeta",
    "ModelType",
    "ActiveModel",
    "AiModel",
    "QuizAnswer",
    "QuizAnswerType",
    "QuizQuestion",
    "QuizQuestionType",
    "ImageClassificationPrediction",
    "VideoPredictionMeta",
    "EndoscopyProcessor",
    "Endoscope",
    "EndoscopeType",
    "TtoQuestionnaire",
    "Frame",
    "ReportFile",
    "Video",
    "LabelVideoSegment",
    "SensitiveMeta",
    "PdfMeta",
    "PdfType",
    "VideoMeta",
    "FFMpegMeta",
    "VideoImportMeta",
    "RawPdfFile",
    "RawVideoFile",
    "VideoSegmentationLabelSet",
]



from django.db import models



class PatientDisease(models.Model):
    patient = models.ForeignKey("Patient", on_delete=models.CASCADE)
    disease = models.ForeignKey("Disease", on_delete=models.CASCADE)
    classification_choices = models.ManyToManyField("DiseaseClassificationChoice")
    start_date = models.DateField(blank=True, null=True)
    end_date = models.DateField(blank=True, null=True)
    numerical_descriptors = models.JSONField(default=dict)
    subcategories = models.JSONField(default=dict)

    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return f"{self.patient} - {self.disease}"
    
    class Meta:
        # unique_together = ('patient', 'disease', 'start_date')
        verbose_name = 'Patient Disease'
        verbose_name_plural = 'Patient Diseases'



from endoreg_db.models import (
    Patient, 
    Center, 
    PatientLabValue, 
    PatientExamination, 
    PatientFinding,
    PatientFindingLocation,
    Finding,
    FindingLocationClassification,
    FindingLocationClassificationChoice,
    FindingMorphologyClassification,
    FindingMorphologyClassificationChoice,
    FindingMorphologyClassificationType,
)
from typing import List
from django.core.management import call_command
from django.test import TestCase
from io import StringIO

from .conf import (
    TEST_CENTER_NAME,
    COLONOSCOPY_FINDING_LOCATION_CLASSIFICATION_NAME,
    TEST_PATIENT_FINDINGS_OUTPUT_PATH,
    TEST_PATIENT_FINDING_LOCATIONS_OUTPUT_PATH,
    TEST_PATIENT_FINDING_MORPHOLOGIES_OUTPUT_PATH,
    TEST_POLYP_MORPHOLOGY_CLASSIFICATION_NAMES
)

from .utils import (
    create_patient_with_colonoscopy,
    create_test_patient
)

class TestGeneratePatientFindings(TestCase):
    def setUp(self):
        out = StringIO()
        call_command("load_gender_data", stdout=out)
        call_command("load_contraindication_data", stdout=out)
        call_command("load_unit_data", stdout=out)
        call_command("load_organ_data", stdout=out)
        call_command("load_examination_data", stdout=out)
        call_command("load_lab_value_data", stdout=out)
        call_command("load_finding_data", stdout=out)
        Center.objects.create(name=TEST_CENTER_NAME)

    def test_create_generic_test_patient_with_colonoscopy_findings(self):
        patient, patient_examination = create_patient_with_colonoscopy(center_name=TEST_CENTER_NAME)
        available_findings = patient_examination.get_available_findings()

        for finding in available_findings:
            patient_examination.create_finding(finding)

        patient_examination.refresh_from_db()
        assert len(patient_examination.get_findings()) == len(available_findings)

        return patient


    def test_create_colonoscopy_patient_findings_locations(self):
        patient, patient_examination = create_patient_with_colonoscopy(center_name=TEST_CENTER_NAME)
        
        # fetch location classification
        location_classification = FindingLocationClassification.objects.get(
            name=COLONOSCOPY_FINDING_LOCATION_CLASSIFICATION_NAME
        )

        available_examination_findings = patient_examination.get_available_findings()
        filtered_findings = location_classification.remove_unavailable_findings(
            available_examination_findings
        )

        location_choices = location_classification.get_choices()
        i = 0
        for finding in filtered_findings:
            for location_choice in location_choices:
                location_choice:FindingLocationClassificationChoice
                subcategories = location_choice.get_subcategories()
                numerical_descriptors = location_choice.get_numerical_descriptors()

                for subcategory, subcategory_dict in subcategories.items():
                    required:bool = subcategory_dict.get("required", False)
                    if required:    
                        subcategory_choices = subcategory_dict.get("choices", [])
                        if subcategory_choices:
                            for choice in subcategory_choices:
                                i+=1
                                patient_finding = patient_examination.create_finding(finding)
                                
                                patient_finding_location = patient_finding.add_location_choice(
                                    location_choice, location_classification
                                )

                                patient_finding_location.set_subcategory(
                                    subcategory,
                                    choice
                                )
                        else: 
                            i+=1
                            patient_finding = patient_examination.create_finding(finding)
                            patient_finding_location = patient_finding.add_location_choice(
                                location_choice, location_classification
                            )

                for numerical_descriptor, numerical_descriptor_dict in numerical_descriptors.items():
                    required:bool = numerical_descriptor_dict.get("required", False)
                    if required:
                        i+=1
                        patient_finding = patient_examination.create_finding(finding)
                        patient_finding_location = patient_finding.add_location_choice(
                            location_choice, location_classification
                        )

                        patient_finding_location.set_random_numerical_descriptor(numerical_descriptor)
                            
        out = ""
        out += f"Created {i} patient findings with locations\n"
        out += f"Tested Patient Findings {filtered_findings}\n"
        out += f"with location classification {location_classification}\n"
        out += f"which contains those location choices {location_choices}\n"
        out += "\n------------\n\n"

        generated_findings = patient_examination.get_findings()

        for patient_finding in generated_findings:
            out += f"Patient Finding: {patient_finding}\n"
            patient_finding_locations = patient_finding.get_locations()
            for _patient_finding_location in patient_finding_locations:
                out += f"Patient Finding Location: {_patient_finding_location}\n"
                out += f"Subcategories: {_patient_finding_location.subcategories}\n"
                out += f"Numerical Descriptors: {_patient_finding_location.numerical_descriptors}\n"
                out += "\n\n"

        with open(TEST_PATIENT_FINDING_LOCATIONS_OUTPUT_PATH, "w") as f:
            f.write(out)


    def test_create_colonoscopy_polyp_morphologies(self):
        FINDING_NAME = "colon_polyp"
        out = ""        

        patient, patient_examination = create_patient_with_colonoscopy(center_name=TEST_CENTER_NAME)

        # get classification types required by colon_polyp
        finding_polyp:Finding = Finding.objects.get(name=FINDING_NAME)

        assert finding_polyp

        req_morphology_classification_types = finding_polyp.get_required_morphology_classification_types()

        morphology_classifications = []
        _types = []
        for classification_name in TEST_POLYP_MORPHOLOGY_CLASSIFICATION_NAMES:
            # print(f"fetching {classification_name}")
            classification = FindingMorphologyClassification.objects.get(
                name=classification_name
            )
            morphology_classifications.append(classification)
            _types.append(classification.classification_type)

        for _ in req_morphology_classification_types:
            assert _ in _types

        i = 0

        morphology_classifications:List[FindingMorphologyClassification]
        for morphology_classification in morphology_classifications:
            morphology_choices = morphology_classification.get_choices()
            for morphology_choice in morphology_choices:
                subcategories= morphology_choice.subcategories
                numerical_descriptors = morphology_choice.numerical_descriptors

                for subcategory, subcategory_dict in subcategories.items():
                    required:bool = subcategory_dict.get("required", False)
                    if required:
                        subcategory_choices = subcategory_dict.get("choices", [])
                        if subcategory_choices:
                            for choice in subcategory_choices:
                                i+=1
                                patient_finding = patient_examination.create_finding(finding_polyp)
                                patient_finding_morphology = patient_finding.add_morphology_choice(
                                    morphology_choice, morphology_classification
                                )
                                patient_finding_morphology.set_subcategory(
                                    subcategory,
                                    choice
                                )
                        else:
                            i+=1
                            patient_finding = patient_examination.create_finding(finding_polyp)
                            patient_finding_morphology = patient_finding.add_morphology_choice(
                                morphology_choice, morphology_classification
                            )

                for numerical_descriptor, numerical_descriptor_dict in numerical_descriptors.items():
                    required:bool = numerical_descriptor_dict.get("required", False)
                    if required:
                        i+=1
                        patient_finding = patient_examination.create_finding(finding_polyp)
                        patient_finding_morphology = patient_finding.add_morphology_choice(
                            morphology_choice, morphology_classification
                        )
                        patient_finding_morphology.set_numerical_descriptor_random(
                            numerical_descriptor
                        )


        out += f"Created {i} patient findings with morphologies\n"
        out += f"Tested Patient Findings for {FINDING_NAME}\n\n----\n\n"

        generated_findings = patient_examination.get_findings()

        for patient_finding in generated_findings:
            out += f"Patient Finding: {patient_finding}\n"
            patient_finding_morphologies = patient_finding.get_morphologies()
            for _patient_finding_morphology in patient_finding_morphologies:
                out += f"Patient Finding Morphology: {_patient_finding_morphology}\n"
                out += f"Subcategories: {_patient_finding_morphology.subcategories}\n"
                out += f"Numerical Descriptors: {_patient_finding_morphology.numerical_descriptors}\n"
                out += "\n\n"
            
        with open(TEST_PATIENT_FINDING_MORPHOLOGIES_OUTPUT_PATH, "w") as f:
            f.write(out)



from django.db import models

class UnitManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class Unit(models.Model):
    objects = UnitManager()

    name = models.CharField(max_length=100) # e.g. "Centimeter"
    name_de = models.CharField(max_length=100, blank=True, null=True) # e.g. "Zentimeter"
    name_en = models.CharField(max_length=100, blank=True, null=True) # e.g. "Centimeter"
    description = models.CharField(max_length=100, blank=True, null=True) # e.g. "centimeters", "milimeters", "inches"
    abbreviation = models.CharField(max_length=25, blank=True, null=True) # e.g. "cm", "mm", "in"
    
    def __str__(self):
        if self.abbreviation:
            return self.abbreviation
        return self.name
    
    def natural_key(self):
        return (self.name,)


import numpy as np
from PIL import Image, ImageOps  # Import the required modules from Pillow

def crop_img(img, crop):
    """
    Crops the image based on the specified dimensions and adds padding to maintain aspect ratio.

    Parameters:
        img: PIL Image object.
        crop: Tuple of (ymin, ymax, xmin, xmax) specifying the crop area.

    Returns:
        PIL Image object that has been cropped and padded as necessary.
    """
    # Convert crop dimensions to Pillow format: left, upper, right, lower
    ymin, ymax, xmin, xmax = crop
    img_cropped = img.crop((xmin, ymin, xmax, ymax))
    
    # Calculate the new size and the required padding
    width, height = img_cropped.size
    delta = width - height
    
    if delta > 0:
        padding = (0, abs(delta) // 2, 0, abs(delta) - abs(delta) // 2)  # (left, top, right, bottom)
    elif delta < 0:
        padding = (abs(delta) // 2, 0, abs(delta) - abs(delta) // 2, 0)
    else:
        padding = (0, 0, 0, 0)
    
    # Pad the image to make it square
    img_padded = ImageOps.expand(img_cropped, padding)
    
    return img_padded


class Cropper:
    def __init__(self):
        pass

    def __call__(self, img, crop=None, scale=None, scale_method=Image.Resampling.LANCZOS):
        """
        Applies cropping and scaling transformations to the input image.

        Parameters:
            img: PIL Image object or numpy array of the image.
            crop: Optional tuple specifying the cropping area (y_min, y_max, x_min, x_max).
            scale: Optional tuple specifying the new size (width, height).
            scale_method: Resampling method used for scaling (default is Image.Resampling.LANCZOS).

        Returns:
            Numpy array of the processed image.
        """
        # Convert numpy array to PIL Image if necessary
        if isinstance(img, np.ndarray):
            img = Image.fromarray(img.astype('uint8'), 'RGB')
        
        if crop is not None:
            img = crop_img(img, crop)
        else:
            raise Exception("Automatic crop detection not implemented yet")
        
        if scale is not None:
            img = img.resize(scale, resample=scale_method)
        
        # Convert PIL Image back to numpy array
        img = np.array(img)
        
        return img


from .examination import Examination
from .examination_type import ExaminationType
from .examination_time import ExaminationTime
from .examination_time_type import ExaminationTimeType
from .examination_indication import(
    ExaminationIndication,
    ExaminationIndicationClassification,
    ExaminationIndicationClassificationChoice
)


from django.core.management.base import BaseCommand
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.apps import apps

class Command(BaseCommand):
    help = 'Sets up default permissions for each model in the app.'

    def handle(self, *args, **options):
        for model in apps.get_models():
            content_type = ContentType.objects.get_for_model(model)
            permissions = ['view', 'add', 'change', 'delete']
            for perm in permissions:
                codename = f'{perm}_{model._meta.model_name}'
                name = f'Can {perm} {model._meta.verbose_name}'
                Permission.objects.get_or_create(
                    codename=codename,
                    defaults={'name': name, 'content_type': content_type},
                )
            self.stdout.write(self.style.SUCCESS(f'Set up permissions for {model._meta.verbose_name}'))



from django.core.management import BaseCommand, call_command
from io import StringIO


class Command(BaseCommand):
    help = "Run all data loading commands in the correct order"

    def add_arguments(self, parser):
        parser.add_argument(
            "--verbose",
            action="store_true",
            help="Display verbose output for all commands",
        )

    def handle(self, *args, **options):
        # verbose = options['verbose']
        verbose = True

        self.stdout.write(self.style.SUCCESS("Populating base db models with data..."))

        out = self.stdout

        # Load Center Data
        call_command("load_center_data", stdout=out, verbose=verbose)
        call_command("load_endoscope_data", stdout=out, verbose=verbose)
        call_command("load_distribution_data", stdout=out, verbose=verbose)

        call_command("load_gender_data", stdout=out, verbose=verbose)
        call_command("load_report_reader_flag_data", stdout=out, verbose=verbose)
        call_command("load_pdf_type_data", stdout=out, verbose=verbose)
        call_command("load_unit_data", stdout=out, verbose=verbose)
        call_command("load_disease_data", stdout=out, verbose=verbose)
        call_command("load_event_data", stdout=out, verbose=verbose)
        call_command("load_organ_data", stdout=out, verbose=verbose)
        call_command("load_contraindication_data", stdout=out, verbose=verbose)
        call_command("load_examination_data", stdout=out, verbose=verbose)
        call_command("load_examination_indication_data", stdout=out, verbose=verbose)
        call_command("load_lab_value_data", stdout=out, verbose=verbose)
        call_command("load_finding_data", stdout=out, verbose=verbose)
        call_command("load_medication_data", stdout=out, verbose=verbose)

        # Load AI Model Data
        call_command("load_ai_model_label_data", stdout=out, verbose=verbose)
        call_command("load_ai_model_data", stdout=out, verbose=verbose)

        self.stdout.write(
            self.style.SUCCESS(  # pylint: disable=no-member
                "All data loading commands executed successfully."
            )
        )

        # # Run the load_network_data command
        # self.stdout.write(self.style.SUCCESS("Running load_network_data..."))
        # call_command('load_network_data', verbose=verbose)
        # Run the load_network_data command
        # self.stdout.write(self.style.SUCCESS("Running load_network_data..."))
        # call_command("load_network_data", verbose=verbose)

        # # Run the load_logging_data command
        # self.stdout.write(self.style.SUCCESS("Running load_logging_data..."))
        # call_command("load_logging_data", verbose=verbose)

        # # Run the load_profession_data command
        # self.stdout.write(self.style.SUCCESS("Running load_profession_data..."))
        # call_command("load_profession_data", verbose=verbose)

        # # run the load_gender_data command
        # self.stdout.write(self.style.SUCCESS("Running load_gender_data..."))
        # call_command("load_gender_data", verbose=verbose)

        # # Run the load_examination_data command
        # self.stdout.write(self.style.SUCCESS("Running load_examination_data..."))
        # call_command("load_examination_data", verbose=verbose)

        # # run the load_finding_data command
        # self.stdout.write(self.style.SUCCESS("Running load_finding_data..."))
        # call_command("load_finding_data", verbose=verbose)

        # # Run the load_disease data command
        # self.stdout.write(self.style.SUCCESS("Running load_disease_data..."))
        # call_command("load_disease_data", verbose=verbose)

        # # # Run the load_disease_classification data command
        # # self.stdout.write(self.style.SUCCESS("Running load_disease_classification_data..."))
        # # call_command('load_disease_classification_data', verbose=verbose)

        # # # Run the load_disease_classification_choices data command
        # # self.stdout.write(self.style.SUCCESS("Running load_disease_classification_choices_data..."))
        # # call_command('load_disease_classification_choices_data', verbose=verbose)

        # # rund the load_user_groups command
        # self.stdout.write(self.style.SUCCESS("Running load_user_groups..."))
        # call_command("load_user_groups", verbose=verbose)

        # # run the load_report_reader_flag command
        # self.stdout.write(self.style.SUCCESS("Running load_report_reader_flag..."))
        # call_command("load_report_reader_flag", verbose=verbose)

        # # Run the load_pdf_type_data command
        # self.stdout.write(self.style.SUCCESS("Running load_pdf_type_data..."))
        # call_command("load_pdf_type_data", verbose=verbose)

        # # Run the load_endoscopy_processor_data command
        # self.stdout.write(
        #     self.style.SUCCESS("Running load_endoscopy_processor_data...")
        # )
        # call_command("load_endoscopy_processor_data", verbose=verbose)

        # # Run the load_endoscope_type_data command
        # self.stdout.write(self.style.SUCCESS("Running load_endoscope_type_data..."))
        # call_command("load_endoscope_type_data", verbose=verbose)

        # # Run the load_unit_data command
        # self.stdout.write(self.style.SUCCESS("Running load_unit_data..."))
        # call_command("load_unit_data", verbose=verbose)

        # # Run the load_lab_value_data command
        # self.stdout.write(self.style.SUCCESS("Running load_lab_value_data..."))
        # call_command("load_lab_value_data", verbose=verbose)

        # # Run the load_information_source command
        # self.stdout.write(self.style.SUCCESS("Running load_information_source..."))
        # call_command("load_information_source", verbose=verbose)

        # # Run the load_center_data command
        # self.stdout.write(self.style.SUCCESS("Running load_center_data..."))
        # call_command("load_center_data", verbose=verbose)

        # # Run the load_label_data command
        # self.stdout.write(self.style.SUCCESS("Running load_label_data..."))
        # call_command("load_label_data", verbose=verbose)

        # # Run the load_ai_model_data command
        # self.stdout.write(self.style.SUCCESS("Running load_ai_model_data..."))
        # call_command("load_ai_model_data", verbose=verbose)

        # # Run the load_event_data command
        # self.stdout.write(self.style.SUCCESS("Running load_event_data..."))
        # call_command("load_event_data", verbose=verbose)

        # # run the load_medication_data command
        # self.stdout.write(self.style.SUCCESS("Running load_medication_data..."))
        # call_command("load_medication_data", verbose=verbose)

        # # # Run the load_medication_indication_type_data command
        # # self.stdout.write(self.style.SUCCESS("Running load_medication_indication_type_data..."))
        # # call_command('load_medication_indication_type_data', verbose=verbose)

        # # # Run the load_medication_intake_time_data command
        # # self.stdout.write(self.style.SUCCESS("Running load_medication_intake_time_data..."))
        # # call_command('load_medication_intake_time_data', verbose=verbose)

        # # # Run the load_medication_schedule_data command
        # # self.stdout.write(self.style.SUCCESS("Running load_medication_schedule_data..."))
        # # call_command('load_medication_schedule_data', verbose=verbose)

        # # # Run the load_medication_indication_data command
        # # self.stdout.write(self.style.SUCCESS("Running load_medication_indication_data..."))
        # # call_command('load_medication_indication_data', verbose=verbose)

        # # Run the load_green_endoscopy_wuerzburg_data command
        # self.stdout.write(
        #     self.style.SUCCESS("Running load_green_endoscopy_wuerzburg_data...")
        # )
        # call_command("load_green_endoscopy_wuerzburg_data", verbose=verbose)

        # # Load G-Play Data
        # # run the load_distribution_data command
        # self.stdout.write(self.style.SUCCESS("Running load_distribution_data..."))
        # call_command("load_distribution_data", verbose=verbose)
        # # Run the load_g_play_data command
        # self.stdout.write(self.style.SUCCESS("Running load_g_play_data..."))
        # call_command("load_g_play_data", verbose=verbose)

        # # Necessary? Migrate? FIXME
        # # # Run the load_active_model_data command
        # # self.stdout.write(self.style.SUCCESS("Running load_active_model_data..."))
        # # call_command('load_active_model_data', verbose=verbose)



from django.db import models
from .patient import Patient

class PatientEvent(models.Model):
    """
    A class representing an event for a patient.

    Attributes:
        patient (Patient): The patient associated with this event.
        date (datetime.date): The date of the event.
        description (str): A description of the event.
    """
    patient = models.ForeignKey(Patient, on_delete=models.CASCADE)
    event = models.ForeignKey("Event", on_delete=models.CASCADE)
    date_start = models.DateField()
    date_end = models.DateField(blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    classification_choice = models.ForeignKey(
        "EventClassificationChoice", on_delete=models.CASCADE, blank=True, null=True
    )

    subcategories = models.JSONField(default=dict)
    numerical_descriptors = models.JSONField(default=dict)

    last_update = models.DateTimeField(auto_now=True)
    
    def __str__(self):
        return str(self.date_start) + ": " + self.event.name
    
    def set_subcategories_from_classification_choice(self):
        """
        Sets the subcategories for this event from the classification choice.
        """
        from endoreg_db.models import EventClassificationChoice
        if self.classification_choice:
            self.classification_choice:EventClassificationChoice
            self.subcategories = self.classification_choice.subcategories
            self.save()
        
        return self.subcategories
    
    def set_numerical_descriptors_from_classification_choice(self):
        """
        Sets the numerical descriptors for this event from the classification choice.
        """
        from endoreg_db.models import EventClassificationChoice
        if self.classification_choice:
            self.classification_choice:EventClassificationChoice
            self.numerical_descriptors = self.classification_choice.numerical_descriptors
            self.save()
        
        return self.numerical_descriptors



from .network_device import NetworkDevice
from .network_device_type import NetworkDeviceType
from .agl_service import AglService

__all__ = [
    "NetworkDevice",
    "NetworkDeviceType",
    "AglService",
]


from django.db import models

class ExaminationTypeManager(models.Manager):
    """
    Manager for ExaminationType with custom query methods.
    """
    def get_by_natural_key(self, name: str) -> "ExaminationType":
        return self.get(name=name)
    
class ExaminationType(models.Model):
    """
    Represents a type of examination.

    Attributes:
        name (str): The unique name of the examination type.
        name_de (str): The German name of the examination type.
        name_en (str): The English name of the examination type.
    """
    objects = ExaminationTypeManager()
    name = models.CharField(max_length=100, unique=True)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)

    def __str__(self) -> str:
        """
        String representation of the examination type.

        Returns:
            str: The name of the examination type.
        """
        return self.name_en or self.name

    def natural_key(self) -> tuple:
        """
        Returns the natural key for the examination type.

        Returns:
            tuple: The natural key consisting of the name.
        """
        return (self.name,)



from .patient_examination import PatientExamination
from .patient_finding import (
    PatientFinding
)
from .patient_finding_location import (
    PatientFindingLocation
)
from .patient_finding_morphology import (
    PatientFindingMorphology
)

from .patient_finding_intervention import (
    PatientFindingIntervention
)

# TODO Migrate to persons/patient

__all__ = [
    "PatientExamination",
    "PatientFinding",
    "PatientFindingLocation",
    "PatientFindingMorphology",
    "PatientFindingIntervention"
]


# import os

# os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.test_settings")

from endoreg_db.models import (
    Unit,
)


from django.core.management import call_command
from django.test import TestCase
from io import StringIO
from .conf import TEST_UNIT_OUTPUT_PATH


class TestCenter(TestCase):
    def setUp(self):
        out = StringIO()
        call_command("load_unit_data", stdout=out)

    def test_units_loaded(self):
        units = Unit.objects.all()
        with open(TEST_UNIT_OUTPUT_PATH, "w") as f:
            for unit in units:
                f.write(f"{unit.name}: {unit}\n\n")



from PIL import Image

def crop_and_insert(image:Image, x, y, h, w, bg_color=(255, 255, 255)):
    """
    Crops a region from an inverted grayscale image and inserts it into a white image of the same size as the original.
    
    Parameters:
    - fp: File path or a file object of the original image.
    - x, y: The top-left coordinates of the rectangle to be cropped.
    - h, w: The height and width of the rectangle to be cropped.
    
    Returns:
    A PIL Image object containing the original image with the specified region replaced.
    """
    # Load the original image
    original_image = image
    
    # Crop the specified region from the inverted image
    crop_rectangle = (x, y, x + w, y + h)
    cropped_content = original_image.crop(crop_rectangle)
    
    # Create a new white image of the same size as the original image
    white_background = Image.new('RGB', original_image.size, bg_color)
    
    # Paste the cropped content onto the white image at the specified location
    white_background.paste(cropped_content, (x, y))
    
    # The final image can be displayed or saved as needed
    return white_background


from pathlib import Path

TEST_REPORT_DIR = Path("test_reports")

if not TEST_REPORT_DIR.exists():
    TEST_REPORT_DIR.mkdir()#

TEST_LX_PERMISSION_OUTPUT_PATH = TEST_REPORT_DIR / "lx_permissions.txt"
TEST_LX_CLIENT_TYPE_OUTPUT_PATH = TEST_REPORT_DIR / "lx_client_types.txt"
TEST_LX_CLIENT_TAGS_OUTPUT_PATH = TEST_REPORT_DIR / "lx_client_tags.txt"

## Client Tags
# - model: endoreg_db.lx_client_tag
#   fields:
#     name: "coloreg"
#     description: "Coloreg related clients"

# - model: endoreg_db.lx_client_tag
#   fields:
#     name: "client"
#     description: "General client"

# - model: endoreg_db.lx_client_tag
#   fields:
#     name: "server"
#     description: "General server"

# - model: endoreg_db.lx_client_tag
#   fields:
#     name: "production"
#     description: "Production environment"

# - model: endoreg_db.lx_client_tag
#   fields:
#     name: "development"
#     description: "Development environment"

# - model: endoreg_db.lx_client_tag
#   fields:
#     name: "test"
#     description: "Test environment"

CLIENT_TAG_NAME_PERMISSIONS_TUPLES = [ #TODO Adapt to endoreg_db
    ("client", ["local_base_db_read", "local_base_db_write"]),
    ("server", ["local_base_db_read", "local_base_db_write"]),
    ("production", ["local_base_db_read", "local_base_db_write"]),
    ("development", ["local_base_db_read", "local_base_db_write"]),
    ("test", ["local_base_db_read", "local_base_db_write"]),
    ("coloreg", ["local_base_db_read", "local_base_db_write"]),
]

CLIENT_TYPE_NAME_ABBREVIATION_TUPLES = [
    ("base-client", "c"),
    ("gpu-client", "gc"),
    ("coloreg-client", "cc"),
    ("base-server", "s"),
    ("gpu-server", "gs"),
    ("storage-server", "ss"),
]

BASE_PERMISSION_NAMES = [
    "local_admin",
    "local_user",
    "local_base_db_read",
    "local_base_db_write",
]


# EndoReg Permission Names
ENDOREG_PERMISSION_NAMES = [
    "endoreg_base_db_read",
    "endoreg_base_db_write",
    "endoreg_base_db_admin",
    "endoreg_base_db_user",

    "endoreg_backup_db_read",
    "endoreg_backup_db_write",
    "endoreg_backup_db_admin",
    "endoreg_backup_db_user",

    "endoreg_test_db_read",
    "endoreg_test_db_write",
    "endoreg_test_db_admin",
    "endoreg_test_db_user",
]



from pathlib import Path

TEST_CENTER_NAME = "Test Center"
TEST_CENTERS_AVAILABLE = [
    "university_hospital_wuerzburg",
    "rbk_stuttgart",
    "bbk_regensburg",
    "uke_hamburg",
    "rkh_ludwigsburg",
    "endoreg_db_demo"
]

TEST_REPORT_DIR = Path("test_reports")

TEST_CENTER_OUTPUT_PATH = TEST_REPORT_DIR / "center.txt"
TEST_ENDOSCOPE_OUTPUT_PATH = TEST_REPORT_DIR / "endoscope.txt"
TEST_GREEN_ENDOSCOPY_OUTPUT_PATH = TEST_REPORT_DIR / "green_endoscopy.txt"




from datetime import date, time
from django.db import models
from .base_classes.abstract_pdf import AbstractPdfFile


class ReportFile(AbstractPdfFile):
    meta = models.JSONField(blank=True, null=True)
    text = models.TextField(blank=True, null=True)
    # sensitive_meta = models.ForeignKey(
    #     "SensitiveMeta",
    #     on_delete=models.SET_NULL,
    #     related_name="report_files",
    #     null=True,
    #     blank=True,
    # )
    # examination = models.ForeignKey(
    #     "PatientExamination",
    #     on_delete=models.SET_NULL,
    #     blank=True,
    #     null=True,
    #     related_name="report_files",
    # )
    # patient = models.ForeignKey(
    #     "Patient", on_delete=models.DO_NOTHING, blank=True, null=True
    # )
    # examiner = models.ForeignKey(
    #     "Examiner", on_delete=models.DO_NOTHING, blank=True, null=True
    # )
    date = models.DateField(blank=True, null=True)
    time = models.TimeField(blank=True, null=True)

    def get_or_create_examiner(self, examiner_first_name, examiner_last_name):
        from ..persons import Examiner

        examiner_center = self.center

        examiner, created = Examiner.objects.get_or_create(
            first_name=examiner_first_name,
            last_name=examiner_last_name,
            center=examiner_center,
        )

        return examiner, created

    def set_examination_date_and_time(self, report_meta=None):
        if not report_meta:
            report_meta = self.meta
        examination_date_str = report_meta["examination_date"]
        examination_time_str = report_meta["examination_time"]

        if examination_date_str:
            # TODO: get django DateField compatible date from string (e.g. "2021-01-01")
            self.date = date.fromisoformat(examination_date_str)
        if examination_time_str:
            # TODO: get django TimeField compatible time from string (e.g. "12:00")
            self.time = time.fromisoformat(examination_time_str)



from endoreg_db.models import (
    Patient, 
    Center, Event,
)
from datetime import datetime
from .utils import create_test_patient
from django.core.management import call_command
from django.test import TestCase
from io import StringIO
from .conf import (
    TEST_CENTER_NAME,
    TEST_PATIENT_EVENT_OUTPUT_PATH
)

class TestGeneratePatientEvents(TestCase):
    def setUp(self):
        out = StringIO()
        call_command("load_gender_data", stdout=out)
        call_command("load_disease_data", stdout=out)
        call_command("load_event_data", stdout=out)
        Center.objects.create(name=TEST_CENTER_NAME)

    def test_patient_event(self):
        patient = create_test_patient(center_name=TEST_CENTER_NAME)

        # get all Events
        events = Event.objects.all()

        out = ""

        for event in events:
            _patient_event = patient.create_event(
                event_name_str=event.name,
                date_start=datetime.now()
            )

            out += f"{event.name}:\n"
            out += f"\t{_patient_event}\n\n"

        with open(TEST_PATIENT_EVENT_OUTPUT_PATH, "w") as f:
            f.write(out)



from django.db import models
    
class CenterWaste(models.Model):
    center = models.ForeignKey("Center", on_delete=models.CASCADE)
    year = models.IntegerField()
    waste = models.ForeignKey("Waste", on_delete=models.CASCADE)
    quantity = models.FloatField()
    unit = models.ForeignKey("Unit", on_delete=models.SET_NULL, null=True)
    emission_factor = models.ForeignKey("EmissionFactor", on_delete=models.SET_NULL, null=True)

    
    def __str__(self):
        result = f"{self.quantity} {self.unit}"
        result += f" -\t{self.waste}, EmissionFactor: {self.emission_factor}\t\t- {self.center} - {self.year}"

        return result



# Generated by Django 5.1.6 on 2025-03-12 05:05

import django.core.files.storage
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('endoreg_db', '0003_alter_frame_image_alter_rawframe_image'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rawvideofile',
            name='file',
            field=models.FileField(storage=django.core.files.storage.FileSystemStorage(location='/home/admin/test-data'), upload_to='db_raw_videos', validators=[django.core.validators.FileExtensionValidator(allowed_extensions=['mp4'])]),
        ),
        migrations.AlterField(
            model_name='video',
            name='file',
            field=models.FileField(storage=django.core.files.storage.FileSystemStorage(location='/home/admin/test-data'), upload_to='db_videos', validators=[django.core.validators.FileExtensionValidator(allowed_extensions=['mp4'])]),
        ),
    ]



from django.db import models
from typing import List

class FindingMorphologyClassificationTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class FindingMorphologyClassificationType(models.Model):
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True)
    name_en = models.CharField(max_length=255, blank=True)
    description = models.TextField(blank=True)
    
    objects = FindingMorphologyClassificationTypeManager()
    
    def natural_key(self):
        return (self.name,)
    
    def __str__(self):
        return self.name
    
    @classmethod
    def get_required_classifications_for_finding(cls, finding):
        """
        Returns all required morphology classification types for a given finding.
        """
        from endoreg_db.models import FindingMorphologyClassificationType
        required_classification_types:List[FindingMorphologyClassificationType] = [
            _ for _ in finding.required_morphology_classification_types.all()
        ]
        return required_classification_types
    
class FindingMorphologyClassificationManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class FindingMorphologyClassification(models.Model):
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True)
    name_en = models.CharField(max_length=255, blank=True)
    description = models.TextField(blank=True)
    classification_type = models.ForeignKey(FindingMorphologyClassificationType, on_delete=models.CASCADE)
    
    objects = FindingMorphologyClassificationManager()
    
    def natural_key(self):
        return (self.name,)
    
    def __str__(self):
        return self.name
    
    def get_choices(self):
        choices: List[FindingMorphologyClassificationChoice] = [_ for _ in self.choices.all()]
        return choices
    
class FindingMorphologyClassificationChoiceManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class FindingMorphologyClassificationChoice(models.Model):
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True)
    name_en = models.CharField(max_length=255, blank=True)
    description = models.TextField(blank=True)
    classification = models.ForeignKey(
        FindingMorphologyClassification, 
        on_delete=models.CASCADE,
        related_name='choices'
    )
    
    subcategories = models.JSONField(
        default = dict
    )

    numerical_descriptors = models.JSONField(
        default = dict
    )

    objects = FindingMorphologyClassificationChoiceManager()
    
    def natural_key(self):
        return (self.name,)
    
    def __str__(self):
        _str = f"{self.name} ({self.classification})"
        return _str

    
        


# Class to represent findings of examinations
from django.db import models
from typing import List

class FindingManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class Finding(models.Model):
    name = models.CharField(max_length=100, unique=True)
    name_de = models.CharField(max_length=100, blank=True, null=True)
    name_en = models.CharField(max_length=100, blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    description_de = models.TextField(blank=True, null=True)
    description_en = models.TextField(blank=True, null=True)
    examinations = models.ManyToManyField('Examination', blank=True, related_name='findings')
    finding_types = models.ManyToManyField('FindingType', blank=True, related_name='findings')

    finding_interventions = models.ManyToManyField(
        'FindingIntervention',
        blank=True,
        related_name='findings'
    )

    causing_finding_interventions = models.ManyToManyField(
        'FindingIntervention',
        blank=True,
        related_name='causing_findings'
    )

    opt_causing_finding_interventions = models.ManyToManyField(
        'FindingIntervention',
        blank=True,
        related_name='opt_causing_findings'
    )

    required_morphology_classification_types = models.ManyToManyField(
        'FindingMorphologyClassificationType',
        blank=True,
        related_name='required_by_findings'
    )

    optional_morphology_classification_types = models.ManyToManyField(
        'FindingMorphologyClassificationType',
        blank=True,
        related_name='optional_for_findings'
    )

    objects = FindingManager()

    def natural_key(self):
        return (self.name,)
    
    def __str__(self):
        return self.name
    
    def get_finding_types(self):
        return self.finding_types.all()
    
    def get_location_classifications(self):
        from endoreg_db.models import FindingLocationClassification
        # FindingLocationClassification is a class with a many-to-many relationship to Finding
        # related name is location_classifications

        location_classifications:FindingLocationClassification = self.location_classifications.all()
        
        return location_classifications 
    
    def get_required_morphology_classification_types(self):
        from endoreg_db.models import FindingMorphologyClassificationType
        finding_morphology_classification_types:List[FindingMorphologyClassificationType] = [
            _ for _ in self.required_morphology_classification_types.all()
        ]
        return finding_morphology_classification_types
    



# Generated by Django 5.1.6 on 2025-03-13 12:00

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('endoreg_db', '0006_frame_extracted_rawframe_extracted'),
    ]

    operations = [
        migrations.RenameField(
            model_name='video',
            old_name='pseudo_patient',
            new_name='patient',
        ),
        migrations.AddField(
            model_name='rawpdffile',
            name='patient',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='raw_pdf_files', to='endoreg_db.patient'),
        ),
    ]



# from django.db import models

# class VideoImportMeta(models.Model):
#     processor = models.ForeignKey('EndoscopyProcessor', on_delete=models.CASCADE)
#     endoscope = models.ForeignKey('Endoscope', on_delete=models.CASCADE, blank=True, null=True)
#     center = models.ForeignKey('Center', on_delete=models.CASCADE)
#     video_anonymized = models.BooleanField(default=False)
#     video_patient_data_detected = models.BooleanField(default=False)
#     outside_detected = models.BooleanField(default=False)
#     patient_data_removed = models.BooleanField(default=False)
#     outside_removed = models.BooleanField(default=False)

#     def __str__(self):
#         result_html = ""

#         result_html += f"Processor: {self.processor.name}<br>"
#         result_html += f"Endoscope: {self.endoscope.name}<br>"
#         result_html += f"Center: {self.center.name}<br>"
#         result_html += f"Video anonymized: {self.video_anonymized}<br>"
#         result_html += f"Video patient data detected: {self.video_patient_data_detected}<br>"
#         result_html += f"Outside detected: {self.outside_detected}<br>"
#         result_html += f"Patient data removed: {self.patient_data_removed}<br>"
#         result_html += f"Outside removed: {self.outside_removed}<br>"
#         return result_html
    



# from endoreg_db.models import (
#     RawVideoFile,Center,VideoMeta
# )
    
# from datetime import datetime

# from django.core.management import call_command
# from django.test import TestCase
# from io import StringIO
# from .conf import (
#     TEST_MULTILABEL_AI_MODEL_IMPORT_PATH,
#     TEST_CENTER_NAME,
#     TEST_VIDEO_DICT,
#     TEST_MULTILABEL_CLASSIFIER_MODEL_PATH,
# )
# from endoreg_db.utils.hashs import (
#     get_hash_string,
# )

# from torch import nn
# import ssl

# ssl._create_default_https_context = ssl._create_unverified_context

# from pathlib import Path

# with open(TEST_MULTILABEL_AI_MODEL_IMPORT_PATH, "w") as f:
#     f.write("Test Multilabel AI Model\n\n")


# AI_SAMPLE_CONFIG = {
#     # mean and std for normalization
#     "mean": (0.45211223, 0.27139644, 0.19264949),
#     "std": (0.31418097, 0.21088019, 0.16059452),
#     # Image Size
#     "size_x": 716,
#     "size_y": 716,
#     # how to wrangle axes of the image before putting them in the network
#     "axes": [2,0,1],  # 2,1,0 for opencv
#     "batchsize": 16,
#     "num_workers": 8, # always 1 for Windows systems # FIXME: fix celery crash if multiprocessing
#     # maybe add sigmoid after prediction?
#     "activation": nn.Sigmoid(),
#     "labels": [
#         'appendix',  'blood',  'diverticule',  'grasper',  'ileocaecalvalve',  'ileum',  'low_quality',  'nbi',  'needle',  'outside',  'polyp',  'snare',  'water_jet',  'wound'
#     ]
# }

# EXTRACTED_TEST_FRAME_DIR = Path("tests/sensitive/frames")

# class TestMultilabelPredictionModel(TestCase):
#     def setUp(self):
#         out = StringIO()
#         call_command("load_gender_data", stdout=out)
#         call_command("load_unit_data", stdout=out)
#         call_command("load_name_data", stdout=out)
#         call_command("load_center_data", stdout=out)
#         call_command("load_endoscope_data", stdout=out)
#         call_command("load_ai_model_data", stdout=out)

#         university_hospital_wuerzburg = Center.objects.get(name=TEST_CENTER_NAME)
#         assert university_hospital_wuerzburg, "Center not found"

#     def test_load_model(self):
#         from endoreg_db.models import RawVideoFile
#         from endoreg_db.models.ai_model.lightning import (
#             MultiLabelClassificationNet,
#             Classifier,
#             InferenceDataset
#         )
        
#         test_video_dict = TEST_VIDEO_DICT["gastroscopy"]
#         center = Center.objects.get(name=TEST_CENTER_NAME)
        
#         model = MultiLabelClassificationNet.load_from_checkpoint(
#             checkpoint_path=TEST_MULTILABEL_CLASSIFIER_MODEL_PATH,
#         )

#         classifier = Classifier(model=model, verbose = True)

#         raw_video = RawVideoFile.create_from_file(
#             file_path=test_video_dict["path"],
#             center_name=center.name,
#             processor_name = test_video_dict["endoscopy_processor"],
#             delete_source=False,
#         )

#         # # generate frames for prediction:
#         extract_frames_result = raw_video.extract_frames(verbose=True)
#         raw_video.update_text_metadata(ocr_frame_fraction=0.001)
#         _r = raw_video.generate_anonymized_frames()
#         anonymized_frame_paths = raw_video.get_frame_paths(anonymized=True)
#         paths_string = [_.as_posix() for _ in anonymized_frame_paths]

#         endo_roi = raw_video.get_endo_roi() # {"x": 0, "y": 0, "width": 1920, "height": 1080}
        
#         crops = [
#             [
#                 endo_roi["y"], 
#                 endo_roi["y"] + endo_roi["height"],
#                 endo_roi["x"],
#                 endo_roi["x"] + endo_roi["width"]
#             ] for _ in paths_string
#         ]

#         ds = InferenceDataset(
#             paths=paths_string,
#             crops=crops,
#             config = AI_SAMPLE_CONFIG
#         )

#         fps = raw_video.get_fps()
#         predictions = classifier.pipe(paths_string, crops)

#         with open(TEST_MULTILABEL_AI_MODEL_IMPORT_PATH, "a") as f:
#             f.write(f"Path example (total: {len(paths_string)}): {paths_string[0]}")

#         readable_predictions = [classifier.readable(p) for p in predictions]

#         result = classifier.post_process_predictions(
#             pred_dicts = readable_predictions,
#             window_size_s = 1,
#             min_seq_len_s = 1.5,
#             fps = fps
#         )

#         for i, _dict in enumerate(result):
#             _keys  = list(_dict.keys())
#             for key in _keys:
#                 # if numpy array
#                 if hasattr(_dict[key], "tolist"):
#                     _dict[key] = _dict[key].tolist()
                
#                 # check if list of tuples
#                 # if so, make sure each tuple has 2 elements and split to two lists (start, stop)
#                 if isinstance(_dict[key], tuple):
#                     if all(len(x) == 2 for x in _dict[key]):
#                         _dict[key] = {
#                             f"{key}_start": [x[0] for x in _dict[key]],
#                             f"{key}_stop": [x[1] for x in _dict[key]]
#                         }
#                         del result[i][key]

#         predictions = result[0]
#         smooth_predictions = result[1]
#         binary_predictions = result[2]
#         raw_sequences = result[3]
#         filtered_sequences = result[4]
#         result_names = [
#             "predictions",
#             "predictions_smooth",
#             "predictions_binary",
#             "raw_sequences",
#             "filtered_sequences"
#         ]
#         # save each file as json
#         import json
#         for i, _dict in enumerate(result):
#             with open(f"results_{result_names[i]}.json", "w") as f:
#                 json.dump(_dict, f, indent=4)



#         with open(TEST_MULTILABEL_AI_MODEL_IMPORT_PATH, "a") as f:
#             f.write(f"Extract Frames Result:\n{extract_frames_result}\n\n")
#             f.write(f"Anonymized Frames Result:\n{_r}\n\n")
#             f.write(f"Predictions:\n{result}\n\n")

       
#         # raw_video.delete_with_file()

#         # for frame_path in anonymized_frame_paths:
#         #     assert frame_path.exists(), f"Anonymized Frame {frame_path} does not exist"



#         # with open(TEST_MULTILABEL_AI_MODEL_IMPORT_PATH, "a") as f:
#         #     f.write(f"Extract Frames Result:\n{extract_frames_result}\n\n")
#         #     f.write(f"Anonymized Frames Result:\n{_r}\n\n")




from .center import Center
from .center_product import CenterProduct
from .center_resource import CenterResource
from .center_waste import CenterWaste

__all__ = [
    'Center',
    'CenterProduct',
    'CenterResource',
    'CenterWaste'
]



from django.db import models
from typing import List

class ExaminationIndicationManager(models.Manager):
    """
    Manager for ExaminationIndication with custom query methods.
    """
    def get_by_natural_key(self, name: str) -> "ExaminationIndication":
        return self.get(name=name)
    
class ExaminationIndication(models.Model):
    """
    Represents an indication for an examination.

    Attributes:
        name (str): The unique name of the indication.
        name_de (str): The German name of the indication.
        name_en (str): The English name of the indication.
        classification (ForeignKey): The classification associated with the indication.
        examination (ForeignKey): The examination associated with the indication.
    """
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)
    classification = models.ForeignKey(
        'ExaminationIndicationClassification', on_delete=models.CASCADE,
        related_name='indications',
        blank=True, null=True
    )
    examination = models.ForeignKey(
        'Examination', on_delete=models.CASCADE,
        related_name='indications',
    )
    
    objects = ExaminationIndicationManager()
    
    def natural_key(self) -> tuple:
        """
        Returns the natural key for the indication.

        Returns:
            tuple: The natural key consisting of the name.
        """
        return (self.name,)

    def __str__(self) -> str:
        """
        String representation of the indication.

        Returns:
            str: The name of the indication.
        """
        return self.name

    def get_choices(self) -> List['ExaminationIndicationClassificationChoice']:
        """
        Retrieves all choices associated with the classification.

        Returns:
            list: A list of classification choices.
        """
        return list(self.classification.choices.all()) if self.classification else []

    def get_examination(self) -> 'Examination':
        """
        Retrieves the associated examination.

        Returns:
            Examination: The associated examination.
        """
        return self.examination

class ExaminationIndicationClassificationManager(models.Manager):
    """
    Manager for ExaminationIndicationClassification with custom query methods.
    """
    def get_by_natural_key(self, name: str) -> "ExaminationIndicationClassification":
        return self.get(name=name)

class ExaminationIndicationClassification(models.Model):
    """
    Represents a classification for examination indications.

    Attributes:
        name (str): The unique name of the classification.
        name_de (str): The German name of the classification.
        name_en (str): The English name of the classification.
    """
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)
    
    
    objects = ExaminationIndicationClassificationManager()
    
    def natural_key(self) -> tuple:
        """
        Returns the natural key for the classification.

        Returns:
            tuple: The natural key consisting of the name.
        """
        return (self.name,)

    def __str__(self) -> str:
        """
        String representation of the classification.

        Returns:
            str: The name of the classification.
        """
        return self.name

    def get_choices(self) -> List['ExaminationIndicationClassificationChoice']:
        """
        Retrieves all choices associated with this classification.

        Returns:
            list: A list of classification choices.
        """
        return list(self.choices.all())

class ExaminationIndicationClassificationChoiceManager(models.Manager):
    """
    Manager for ExaminationIndicationClassificationChoice with custom query methods.
    """
    def get_by_natural_key(self, name: str) -> "ExaminationIndicationClassificationChoice":
        return self.get(name=name)

class ExaminationIndicationClassificationChoice(models.Model):
    """
    Represents a choice within an examination indication classification.

    Attributes:
        name (str): The unique name of the choice.
        name_de (str): The German name of the choice.
        name_en (str): The English name of the choice.
        subcategories (JSONField): Subcategories associated with the choice.
        numerical_descriptors (JSONField): Numerical descriptors for the choice.
        classification (ForeignKey): The classification to which this choice belongs.
    """
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)
    subcategories = models.JSONField(default=dict)
    numerical_descriptors = models.JSONField(default=dict)
    classification = models.ForeignKey(
        ExaminationIndicationClassification, on_delete=models.CASCADE,
        related_name='choices'
    )

    objects = ExaminationIndicationClassificationChoiceManager()

    def natural_key(self) -> tuple:
        """
        Returns the natural key for the classification choice.

        Returns:
            tuple: The natural key consisting of the name.
        """
        return (self.name,)

    def __str__(self) -> str:
        """
        String representation of the classification choice.

        Returns:
            str: The name of the classification choice.
        """
        return self.name


from endoreg_db.models import (
    Patient, 
    Center, 
)

from .utils import create_test_patient

from django.core.management import call_command
from django.test import TestCase
from io import StringIO
from .conf import (
    TEST_CENTER_NAME,
    TEST_PATIENT_OUTPUT_PATH
)

class TestGeneratePatient(TestCase):
    def setUp(self):
        out = StringIO()
        call_command("load_gender_data", stdout=out)
        call_command("load_unit_data", stdout=out)
        call_command("load_examination_data", stdout=out)
        Center.objects.create(name=TEST_CENTER_NAME)

        # Delete existing test_patient_examination_output_path
        with open(TEST_PATIENT_OUTPUT_PATH, "w") as f:
            f.write("")

    def test_create_patient(self):
        patient = create_test_patient()

        report = "Test Patient created successfully: \n"
        report += f"{patient}\n"

        # Append to test_patient_examination_output_path
        with open(TEST_PATIENT_OUTPUT_PATH, "a") as f:
            f.write(report)

        assert patient

    def test_create_100_patients(self):
        # delete all patients
        Patient.objects.all().delete()

        # Create 100 patients
        for i in range(100):
            _patient = create_test_patient()

        # calculate mean, min, max and std of patient.age
        patients = Patient.objects.all()
        ages = [patient.age() for patient in patients]

        mean_age = sum(ages) / len(ages)
        min_age = min(ages)
        max_age = max(ages)

        std_age = (sum([(age - mean_age)**2 for age in ages]) / len(ages))**0.5

        report = "100 Test Patients created successfully: \n"
        report += f"Mean age: {mean_age}\n"
        report += f"Min age: {min_age}\n"
        report += f"Max age: {max_age}\n"
        report += f"Std age: {std_age}\n"

        # Append to test_patient_examination_output_path
        with open(TEST_PATIENT_OUTPUT_PATH, "a") as f:
            f.write(report)

